{-----------------------------------------------------------------------------
 Unit Name: ConstTypesAndGlobals
 Project Name : Maps
 Author:    Angel Montesinos-Amilibia
 January 2009
 Purpose: Several auxiliary maps
-----------------------------------------------------------------------------}
{

MapR2toR2, workshop for smooth maps from R^2 to R^2
Copyright (C) 2008-2009 Angel Montesinos-Amilibia,
   University of Valencia, Spain


This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

unit Maps;

interface
  uses
    Windows, Classes, 
    OCComplexFunct, OCFunctDDDoubleSafe, FMapR2R2, ConstTypesAndGlobals;

var
   theF: R2ToR2Function; {the function that defines the map}
   theFdiff1: Diff1R2ToR2Procedure; {its differential}
   theJac: R2ToR1Function;  {its Jacobian}
   theJacGrad: R2ToR2Function;  {the gradient of the Jacobian}
   DomPanelToR2: PtBToR2Function; {map from the pixels of the domain panel
                                   to the domain of 'theF' or to a sphere
                                   when the stereographic option for the domain
                                   has been chosen}
   DomR3ToPanel: R3ToR2Function; {from the domain space, directly or as
                                  stereo-projected into the sphere, to the domain
                                  panel, according with the current option}
   ImgR3ToPanel: R3ToR2Function;  {the same for the image space with respect to the
                                   image panel}

   DomToImgR3: R2ToR3Function;    {from the points of the domain to the image space,
                                   directly or as stereo-projected into a sphere,
                                   according with the current option}
   ImgPanelToR2: PtBToR2Function;  {from the pixels of the image panel to the image
                                   space, directly or as represented in the sphere,
                                   according with the current option}

   panel2domSc, panel2imgSc: Vector2;
   domEffRct, imgEffRct: TRect;   {when the "preserve ratio" as been chosen, the
                                   useful part of the panels are sometimes reduced
                                   to some rectangle, the "effective" rectangle}
   domToDomPanel: R2ToBPtFunction;
   imgToImgPanel: R2ToBPtFunction;

   procedure UpdateMaps;
     {Assigns the above functions to specificly defined funtions according with
     the current options}

   function DomToDomR3(const par: Vector2): Vector3; {from the domain to R^3: it is
                   the stereographic projection or the inclusion to the plane z=0,
                   according with the current option}
   procedure DomToImgR3diff1(const par: Vector2; var y, yu, yv: Vector3);
    {the differential of the map 'DomToImgR3'}

   function DomToImgR3St(const par: vector2): vector3; {From a point 'par' in
        the domain to theF(par), and then by stereo-projection to the sphere. It
        is equal to 'DomToImgR3' when the stereo option has been chosen for
        the image}

   function InDom(var domPt, imgPt: Vector2; x,y: Integer; var back: Integer): Boolean;
   {Let [x,y] be a pixel in the domain panel. If the flat option is chosen,
    then the routine returns in 'domPt' the point in the domain that is
    represented by that pixel. If the stereo option is chosen for the domain,
    then there are none, one, two points of the sphere --and therefore two
    points of the domain R^2 space-- represented by that pixel. In this case
    it returns in 'domPt' the point of the domain R^2 whose image in the sphere
    is nearest to the viewer. In both options, the result is true if the point
    'domPt' is inside the limits of the domain.}


   function MoveStereoSettings(var sts: StereoSettings; lastPt, curPt: TPoint;
        Shift: TShiftState): Boolean;
   {In response to a mouse movement, rotate the viewpoint, of change sphere radius,
    or move the center of the sphere along the plane r^3 of zero third coordinate}

   procedure SetCoordinateChangesDom(rc: TRect; minC, maxC: Vector2;
        preserveRatio: Boolean);
   {set coefficients for the functions 'FlatDom3DToPanel' and its inverse}

   procedure SetCoordinateChangesImg(rc: TRect; minC, maxC: Vector2;
        preserveRatio: Boolean);
   {set coefficients for the functions 'FlatImg3DToPanel' and its inverse}


implementation
uses
  Math;

var

  minDom, maxDom, minImg, maxImg, dom2panelSc, img2panelSc: Vector2;
  domCtr, imgCtr: TPoint;

function StProjection(const sts: StereoSettings; p: Vector2): Vector3;
{Gives the stereographic projection of the vector 'p' into the sphere given by
 sts}
var
  a, aux: Double;
begin
  a:= sqr(p[1]-sts.c[1])+ sqr(p[2]-sts.c[2]);
  aux:= sts.r/(sts.r2 + a);
  result[3]:= aux*(sts.r2 - a);
  aux:= 2 * sts.r * aux;
  result[1]:= aux*(p[1]-sts.c[1]);
  result[2]:= aux*(p[2]-sts.c[2]);
end;

function DomToDomPanelFlat(const x: Vector2; var back: Boolean): TPoint;
begin
  result.x:= round((x[1]-minDom[1])*dom2panelSc[1] + domEffRct.left);
  result.y:= round((-x[2]+ maxDom[2])*dom2panelSc[2] + domEffRct.top);
  back:= False;
end;

function DomToDomPanelSt(const x: Vector2; var back: Boolean): TPoint;
var
  z: Vector3;
begin
  z:= StProjection(stsDom, x);
  result.x:= round(stsDom.sc*InnerProd(z, stsDom.h)+ domCtr.x);
  result.y:= round(stsDom.sc*InnerProd(z, stsDom.v)+ domCtr.y);
  back:=  InnerProd(z,stsDom.n) > 0;
end;

function DomToDomR3(const par: vector2): vector3;
{From a point 'par' in the domain to the sphere that represents the domain}
begin
  if mapR2R2.modeStDomain then
    result:= StProjection(stsDom,par)
  else begin
    result[1]:= par[1];
    result[2]:= par[2];
    result[3]:= 0;
  end
end;

procedure DomToImgR3diff1(const par: Vector2; var y, yu, yv: Vector3);
const
  incr= 1.0e-6;
  factor= 5.0e5;
var
  p: Vector2;
  yb, yf: Vector3;
  i: Integer;
begin
  y:= DomToImgR3(par);
  p[1]:= par[1] + incr;
  p[2]:= par[2];
  yf:= DomToImgR3(p);
  p[1]:= par[1] - incr;
  yb:= DomToImgR3(p);
  for i:= 1 to 3 do
    yu[i]:= factor*(yf[i]-yb[i]);
  p[1]:= par[1] ;
  p[2]:= par[2] + incr;
  yf:= DomToImgR3(p);
  p[2]:= par[2] - incr;
  yb:= DomToImgR3(p);
  for i:= 1 to 3 do
    yv[i]:= factor*(yf[i]-yb[i]);
end;

function DomToImgR3Flat(const par: vector2): vector3;
{From a point 'par' in the domain to theF(par) that represents the domain,
 and then to the sphere}
var
  y: Vector2;
begin
  y:= theF(par);
  result[1]:= y[1];
  result[2]:= y[2];
  result[3]:= 0;
end;

function DomToImgR3St(const par: vector2): vector3;
{From a point 'par' in the domain to theF(par) that represents the domain,
 and then to the sphere}
var
  y: Vector2;
begin
  y:= theF(par);
  result:= StProjection(stsImg,y);
end;

function DomR3ToPanelFlat(const x: Vector3; var back: Boolean): Vector2;
{from the domain space as included in R^3 to the domain panel}
begin
  result[1]:= (x[1]-minDom[1])*dom2panelSc[1] + domEffRct.left;
  result[2]:= (-x[2]+ maxDom[2])*dom2panelSc[2] + domEffRct.top;
  back:= False;
end;

function DomR3ToPanelSt(const x: Vector3; var back: Boolean): Vector2;
{from the domain space, as stereo-projected into the sphere, to the domain panel}
begin
  result[1]:= stsDom.sc*InnerProd(x, stsDom.h)+ domCtr.x;
  result[2]:= stsDom.sc*InnerProd(x, stsDom.v)+ domCtr.y;
  back:=  InnerProd(x,stsDom.n) > 0;
end;

function DomPanelToR2Flat(const x,y: Integer; back: Boolean): Vector2;
{from the pixel[i,j] in the domain panel to the domain space}
begin
  result[1]:= (x - domEffRct.left)*panel2domSc[1] + minDom[1];
  result[2]:= (domEffRct.top - y)*panel2domSc[2] + maxDom[2];
end;

function DomPanelToR2St(const x,y: Integer; back: Boolean): Vector2;
{Given a pixel of coordinates 'x,y' in the domain panel, where it is
represented the domain in stereographic manner, and the value of 'back'
to distinguish between the two points of the sphere that project upon
that pixel, return the corresponding point in the domain. If the pixel
is out of the circle on which the sphere projects, result[1]:= kTheError
and result[2]:= 0.}
var
  aux, u, v: Double;
  z: Vector3;
  i: Integer;
  rslt: Boolean;
begin
  u:= (x - stsDom.ctr.x)/stsDom.sc;
  v:= (y - stsDom.ctr.y)/stsDom.sc;
  aux:= sqr(stsDom.r) - sqr(u) - sqr(v);
  rslt:= ( aux > 0 );
  if rslt then begin
    aux:= (2*ord(back)-1)*sqrt(aux);
    for i:= 1 to 3 do
      z[i]:= u*stsDom.h[i] + v*stsDom.v[i]+ aux*stsDom.n[i];
    aux:= stsDom.r + z[3];
    rslt:= ( aux > 0 );
    if rslt then begin
      result[1]:= stsDom.c[1]+ stsDom.r*z[1]/aux;
      result[2]:= stsDom.c[2]+ stsDom.r*z[2]/aux;
    end
  end;
  if (not rslt) then begin
    result[1]:= kTheError;
    result[2]:= 0;
  end;
end;

function ImgToImgPanelFlat(const x: Vector2; var back: Boolean): TPoint;
begin
  result.x:= round((x[1]-minImg[1])*img2panelSc[1] + imgEffRct.left);
  result.y:= round((-x[2]+ maxImg[2])*img2panelSc[2] + imgEffRct.top);
  back:= False;
end;

function ImgToImgPanelSt(const x: Vector2; var back: Boolean): TPoint;
  var
    z: Vector3;
begin
  z:= StProjection(stsImg, x);
  result.x:= round(stsImg.sc*InnerProd(z, stsImg.h)+ imgCtr.x);
  result.y:= round(stsImg.sc*InnerProd(z, stsImg.v)+ imgCtr.y);
  back:=  InnerProd(z,stsImg.n) > 0;
end;

function ImgR3ToPanelFlat(const x: Vector3; var back: Boolean): Vector2;
{from the image space as included in R^3 to the domain panel}
begin
  result[1]:= (x[1] - minImg[1])*img2panelSc[1] + imgEffRct.left;
  result[2]:= (maxImg[2] - x[2])*img2panelSc[2] + imgEffRct.top;
  back:= False;
end;

function ImgR3ToPanelSt(const x: Vector3; var back: Boolean): Vector2;
{from the image space, as stereo-projected into the sphere, to the image panel}
begin
  result[1]:= stsImg.sc*InnerProd(x, stsImg.h)+ imgCtr.x;
  result[2]:= stsImg.sc*InnerProd(x, stsImg.v)+ imgCtr.y;
  back:=  InnerProd(x,stsImg.n) > 0;
end;

function ImgPanelToR2Flat(const x,y: Integer; back: Boolean): Vector2;
{from the pixel[i,j] in the image panel to the image space}
begin
  result[1]:= (x - imgEffRct.left)*panel2imgSc[1] + minImg[1];
  result[2]:= (imgEffRct.top - y)*panel2imgSc[2] + maxImg[2];
end;

function ImgPanelToR2St(const x,y: Integer; back: Boolean): Vector2;
{Given a pixel of coordinates 'x,y' in the image panel, where it is
represented the image space in stereographic manner, and the value of 'back'
to distinguish between the two points of the sphere that project upon
that pixel, return the corresponding point in the image space. If the pixel
is out of the circle on which the sphere projects, result[1]:= kTheError and
result[2]:= 0.}
var
  aux, u, v: Double;
  z: Vector3;
  i: Integer;
  rslt: Boolean;
begin
  u:= (x - stsImg.ctr.x)/stsImg.sc;
  v:= (y - stsImg.ctr.y)/stsImg.sc;
  aux:= sqr(stsImg.r) - sqr(u) - sqr(v);
  rslt:= ( aux > 0 );
  if rslt then begin
    aux:= (2*ord(back)-1)*sqrt(aux);
    for i:= 1 to 3 do
      z[i]:= u*stsImg.h[i] + v*stsImg.v[i]+ aux*stsImg.n[i];
    aux:= stsImg.r + z[3];
    rslt:= ( aux > 0 );
    if rslt then begin
      result[1]:= stsImg.c[1]+ stsImg.r*z[1]/aux;
      result[2]:= stsImg.c[2]+ stsImg.r*z[2]/aux;
    end
  end;
  if (not rslt) then begin
    result[1]:= kTheError;
    result[2]:= 0;
  end;
end;

function InDom(var domPt, imgPt: Vector2; x,y: Integer; var back: Integer): Boolean;
   {Let [x,y] be a pixel in the domain panel. If the flat option is chosen,
    then the routine returns in 'domPt' the point in the domain that is
    represented by that pixel. If the stereo option is chosen for the domain,
    then there are none, one, two points of the sphere --and therefore two
    points of the domain R^2 space-- represented by that pixel. In this case
    it returns in 'domPt' the point of the domain R^2 whose image in the sphere
    is nearest to the viewer. In both options, the result is true if the point
    'domPt' is inside the limits of the domain.}
  var
    v1, v2: Vector2;
begin
  if mapR2R2.modeStDomain then begin
    case back of
      0: v1:= DomPanelToR2(x,y,False);
      1: v1:= DomPanelToR2(x,y,True);
      else begin
        v1:= DomPanelToR2(x,y,False);
        v2:= DomPanelToR2(x,y,True);
      end;
    end;
    if back in [0,1] then begin
      result:= (v1[1]<> kTheError) and (v1[1]>= minDom[1]) and (v1[1] <= maxDom[1]) and
                 (v1[2]>= minDom[2]) and (v1[2] <= maxDom[2]);
      domPt:= v1;
    end
    else begin
      result:= (v1[1]<> kTheError) and (v1[1]>= minDom[1]) and (v1[1] <= maxDom[1]) and
                 (v1[2]>= minDom[2]) and (v1[2] <= maxDom[2]);
      domPt:= v1;
      back:= 0;
      if not result then begin
        result:= (v2[1]<> kTheError) and (v2[1]>= minDom[1]) and (v2[1] <= maxDom[1]) and
                 (v2[2]>= minDom[2]) and (v2[2] <= maxDom[2]);
        domPt:= v2;
        back:= 1;
      end;
    end
  end
  else begin
    domPt:= DomPanelToR2Flat(x,y,True);
    result:= (domPt[1]>= minDom[1]) and (domPt[1] <= maxDom[1]) and
             (domPt[2]>= minDom[2]) and (domPt[2] <= maxDom[2]);
    back:= 0;
  end;
  imgPt:= theF(domPt);
end;

function DefinedRotVectors(sts: StereoSettings; p1, p2: TPoint; var e1,e2: Vector3): Boolean;
{We consider that the screen in a plane by the origin of R^3 and that
 (vH, vV) is an orthonormal basis of that plane, so that the horizontal
 screen lines are parallel to vH and the vertical ones parallel to vV. Also
 vN is a unit vector orthogonal to that plane. For a point p in the screen
 we get the inverse stereographic projection on a sphere centered at the origin
 with radius 'sts.rEff'. Then we normalize the resulting vector. This we do
 with p1 and p2. The resulting points will define a rotation of R^3 that
 will be used to rotate the sphere.}
var
  t, x, y, d2, r, r2, sr, ph, pn, pv: double;
  i: Integer;
begin
  r:= sts.r*sts.sc;
  r2:= sqr(r);
  x:= p1.x- sts.ctr.x;  {components of p1 relative to the panel center}
  y:= p1.y- sts.ctr.y;
  d2:= sqr(x)+ sqr(y);
  t:= 2*r2/(r2 + d2);

  {components of inverse stereographic projection of 'p1' in the basis
   (vHImg, vVImg, vNImg)}
  ph:= x*t;
  pv:= y*t;
  pn:= r*(1-t);

  for i:= 1 to 3 do
    e1[i]:= ph*sts.h[i]+ pv*sts.v[i]+ pn*sts.n[i];  {value of the inverse st. proj of p1}
  sr:= sqrt(sqr(e1[1])+ sqr(e1[2])+sqr(e1[3]));
  result:= (sr > 1.0e-5);
  if result then begin
    for i:= 1 to 3 do
      e1[i]:= e1[i]/sr;  {e1 becomes normalized, and we do the same with 'p2'}
    x:= p2.x- sts.ctr.x;
    y:= p2.y- sts.ctr.y;
    d2:= sqr(x)+ sqr(y);
    t:= 2*r2/(r2 + d2);
    ph:= x*t;
    pv:= y*t;
    pn:= r*(1-t);
    for i:= 1 to 3 do
      e2[i]:= ph*sts.h[i]+ pv*sts.v[i]+ pn*sts.n[i];
    sr:= sqrt(sqr(e2[1])+ sqr(e2[2])+sqr(e2[3]));
    result:= (sr > 1.0e-5);
    if result then begin
      for i:= 1 to 3 do
        e2[i]:= e2[i]/sr;
    end;
  end;
end;

procedure ComputeNewViewBasis(var sts: StereoSettings; lastPt, curPt: TPoint);
{Mouse was dragged from 'lastPt' to 'curPt'. Those points, by an
 inverse stereographic projection give two unit vectors that define
 a rotation around the axis orthogonal to both vectors passing by
 the origin, rotation that sends the first unit vector to the other.
 Here we use the decomposition of a rotation into two reflections,
 as was shown to me by Dick Palais, from Lehigh University. Of course,
 all this is done in the current basis of the plane where the projection
 is made, and we obtain the rotation by rotating the basis in the opposite
 sense}
var
  nn, twoVe2, vb, e2b: Double;
  e1, e2, b: Vector3;
  i: Integer;
begin
  if DefinedRotVectors(sts, curPt, lastPt,  e1, e2) then begin
    for i := 1 to 3 do
      b[i] := e1[i] + e2[i];
      nn := Norm(b);
      for i := 1 to 3 do
        b[i] := b[i] / nn;
      e2b := InnerProd(e2,b);
      twoVe2 := 2 * InnerProd(e2,sts.h);
      vb := InnerProd(b,sts.h);
      vb := 2 * (twoVe2 * e2b - vb);
      for i := 1 to 3 do
        sts.h[i] := vb * b[i] - twoVe2 * e2[i] + sts.h[i];
      twoVe2 := 2 * InnerProd(e2,sts.v);
      vb := InnerProd(b,sts.v);
      vb := 2 * (twoVe2 * e2b - vb);
      for i := 1 to 3 do
        sts.v[i] := vb * b[i] - twoVe2 * e2[i] + sts.v[i];
      twoVe2 := 2 * InnerProd(e2,sts.n);
      vb := InnerProd(b,sts.n);
      vb := 2 * (twoVe2 * e2b - vb);
      for i := 1 to 3 do
        sts.n[i] := vb * b[i] - twoVe2 * e2[i] + sts.n[i];
  end;
end;

function MoveStereoSettings(var sts: StereoSettings; lastPt, curPt: TPoint;
        Shift: TShiftState): Boolean;
{In response to a mouse movement, rotate the viewpoint, of change sphere radius,
or move the center of the sphere along the plane r^3 of zero third coordinate}
var
  nn, scale, a: Double;
  displ: Vector2;
begin
  if ssShift in Shift then begin  {change sphere radius}
    result:= ( curPt.Y <> lastPt.y );
    if result then begin
      scale:= max(0.1,(1 + (curPt.Y - lastPt.y) / 200));
      scale:= min(10, scale);
      a:= Max(1.0e-3, sts.r*scale); {Minimum radius allowed is 0.001}
      a:= Min(1.0e3,a);                {Maximum radius allowed is 1000}
      result := abs(a-sts.r)> 1.0e-4;
      if result then begin
        sts.sc:= sts.sc*sts.r/a;
        sts.r:= a;
        sts.r2:= sqr(sts.r);
      end;
    end;
  end
  else if ssCtrl  in Shift then begin {move sphere center}
    displ[1]:= lastPt.x- curPt.x;
    displ[2]:= lastPt.y- curPt.y;
    nn:= hypot(displ[1],displ[2]);
    result:= nn > 0;
    if result then begin
      {scale the movement to the relative size of sphere}
      sts.c[1]:= sts.c[1]+ displ[1]/sts.sc;
      sts.c[2]:= sts.c[2]+ displ[2]/sts.sc;
    end;
  end
  else begin  {rotate the viewpoint of the sphere}
    result:= (lastPt.x <> curPt.x) or (lastPt.y <> curPt.y);
    if result then
      ComputeNewViewBasis(sts, curPt, lastPt);
  end;
end;


function TheCF(const par: Vector2): vector2;
{wrap for the values of the complex function}
begin
  theComplexF.V(result,[MapR2R2.pComplex,par]);
end;


procedure TheCFDiff1(const par: vector2; var x,xu,xv: Vector2);
{gives the values of the complex function and its derivatives using
 four-points divided differences}
  const
    incr =  0.0002;
    doubleIncr= 2* incr;
    twelveIncr= 12*incr;
    twoSqrIncr= 2*incr*incr;
    twelveSqrIncr= 6*twoSqrIncr;
  var
    i: Integer;
    varPar, x1, x2, x3, x4: Vector2;
begin
  x:= theCF(par);
  varPar:= par;

  varpar[1]:= par[1] + incr;
  x1:= theCF(varPar);

  varPar[1]:= par[1] + doubleIncr;
  x2:= theCF(varPar);

  varpar[1]:= par[1] - incr;
  x3:= theCF(varPar);

  varPar[1]:= par[1] - doubleIncr;
  x4:= theCF(varPar);
  for i:= 1 to 2 do
    xu[i]:= (8*(x1[i]- x3[i]) + x4[i] - x2[i])/twelveIncr;

  varpar[1]:= par[1] ;
  varpar[2]:= par[2]+ + incr;
  x1:= theCF(varPar);

  varPar[2]:= par[2] + doubleIncr;
  x2:= theCF(varPar);

  varpar[2]:= par[2] - incr;
  x3:= theCF(varPar);

  varPar[2]:= par[2] - doubleIncr;
  x4:= theCF(varPar);
  for i:= 1 to 2 do
    xv[i]:= (8*(x1[i]- x3[i]) + x4[i] - x2[i])/twelveIncr;
end;

function theCFJac(const par: Vector2): Double;
{Jacobian of the complex function}
  var
    x, xu, xv: Vector2;
begin
  theCFDiff1(par,x,xu,xv);
  result:= xu[1]*xv[2]- xu[2]*xv[1];
end;

function TheCFJacGrad(const par : vector2): Vector2;
{Jacobian of the real map}
  const
    incr =  0.0002;
    doubleIncr= 2* incr;
    twelveIncr= 12*incr;
    twoSqrIncr= 2*incr*incr;
    twelveSqrIncr= 6*twoSqrIncr;
  var
    i: Integer;
    varPar: Vector2;
    x, x1, x2, x3, x4: Double;
begin
  x:= theCFJac(par);
  varPar:= par;

  varpar[1]:= par[1] + incr;
  x1:= theCFJac(varPar);

  varPar[1]:= par[1] + doubleIncr;
  x2:= theCFJac(varPar);

  varpar[1]:= par[1] - incr;
  x3:= theCFJac(varPar);

  varPar[1]:= par[1] - doubleIncr;
  x4:= theCFJac(varPar);

  result[1]:= (8*(x1- x3) + x4 - x2)/twelveIncr;

  varpar[1]:= par[1] ;
  varpar[2]:= par[2]+ incr;
  x1:= theCFJac(varPar);

  varPar[2]:= par[2] + doubleIncr;
  x2:= theCFJac(varPar);

  varpar[2]:= par[2] - incr;
  x3:= theCFJac(varPar);

  varPar[2]:= par[2] - doubleIncr;
  x4:= theCFJac(varPar);

  result[2]:= (8*(x1- x3) + x4 - x2)/twelveIncr;
end;


function TheFReal(const par: Vector2) : vector2;
{Wrap for the values of the real function}
begin
  result[1]:= theF1.V(par);
  result[2]:= theF2.V(par);
end;

function TheFRealJac(const p : vector2) : double;
{Jacobian of the real map}
var
  fd: FirstDeriv;
  xu, xv, yu, yv : double;
begin
  fd:= theF1.FirstD([p[1], p[2],1,0]);
  xu:= fd.dfa0;
  fd:= theF1.FirstD([p[1], p[2],0,1]);
  xv:= fd.dfa0;
  fd:= theF2.FirstD([p[1], p[2],1,0]);
  yu:= fd.dfa0;
  fd:= theF2.FirstD([p[1], p[2],0,1]);
  yv:= fd.dfa0;
  result := xu * yv - xv * yu;
end;

function TheFRealJacGrad(const p : vector2) : Vector2;
{Gradient of the Jacobian of the real map}
var
  sd: SecondDeriv;
  xu, xv, yu, yv, xuu, xuv, xvv, yuu, yvv, yuv: Double;
begin
  sd:= theF1.SecondD([p[1], p[2],1,0,0,0]);
  xu:= sd.dfa0;
  xuu:= sd.ddfa0;
  sd:= theF1.SecondD([p[1], p[2],0,1,0,0]);
  xv:= sd.dfa0;
  xvv:= sd.ddfa0;
  sd:= theF1.SecondD([p[1], p[2],1,1,0,0]);
  xuv:= 0.5*(sd.ddfa0 - xuu - xvv);

  sd:= theF2.SecondD([p[1], p[2],1,0,0,0]);
  yu:= sd.dfa0;
  yuu:= sd.ddfa0;
  sd:= theF2.SecondD([p[1], p[2],0,1,0,0]);
  yv:= sd.dfa0;
  yvv:= sd.ddfa0;
  sd:= theF2.SecondD([p[1], p[2],1,1,0,0]);
  yuv:= 0.5*(sd.ddfa0 - yuu - yvv);
  result[1]:= xuu*yv + xu*yuv - xuv*yu - xv*yuu;
  result[2]:= xuv*yv + xu*yvv - xvv*yu - xv*yuv;
end;

procedure TheFRealDiff1(const par: Vector2; var x, xu, xv: Vector2);
{Value and first partial derivatives of the real map}
var
  fd: FirstDeriv;
begin
  fd:= theF1.FirstD([par[1], par[2],1,0]);
  x[1]:= fd.fa0;
  xu[1]:= fd.dfa0;

  fd:= theF1.FirstD([par[1], par[2],0,1]);
  xv[1]:= fd.dfa0;

  fd:= theF2.FirstD([par[1], par[2],1,0]);
  x[2]:= fd.fa0;
  xu[2]:= fd.dfa0;

  fd:= theF2.FirstD([par[1], par[2],0,1]);
  xv[2]:= fd.dfa0;
end;

procedure SetCoordinateChangesDom(rc: TRect; minC, maxC: Vector2; preserveRatio: Boolean);
{Sets coefficients for the functions 'FlatDom3DToPanel' and its inverse}
var
  i,w, cut: Integer;
  edges: Vector2;
begin
  w:= rc.right - rc.left -1;
  domCtr.x:= (rc.right + rc.left) div 2;
  domCtr.y:= (rc.top + rc.bottom) div 2;
  minDom:= minC;
  maxDom:= maxC;
  if MapR2R2.modeStDomain then begin
    stsDom.sc:= 0.45*(w+1)/stsDom.r;
    stsDom.ctr:= domCtr;
  end;
  for i:= 1 to 2 do begin
    edges[i]:= maxDom[i]- minDom[i];
    dom2panelSc[i]:=  (w+1)/edges[i];
  end;
  if (not MapR2R2.modeStDomain) and preserveRatio then begin {the zoom factor must be equal in both directions.
      So, we take the smaller of the two and reduce conveniently the
      rectangle where the drawing is to be done}
    if dom2panelSc[1] < dom2panelSc[2] then begin
      cut:= round(0.5*(w - edges[2]*dom2panelSc[1]));
      dom2panelSc[2]:= dom2panelSc[1];
      SetRect(domEffRct, rc.left, rc.Top + cut, rc.right, rc.bottom - cut);
    end
    else begin
      dom2panelSc[1]:= dom2panelSc[2];
      cut:= round(0.5*(w - edges[1]*dom2panelSc[2]));
      SetRect(domEffRct, rc.left+ cut, rc.Top, rc.right - cut, rc.bottom);
    end;
  end
  else
    domEffRct:= rc;
  for i:= 1 to 2 do
    panel2DomSc[i]:= 1/dom2panelSc[i];
end;

procedure SetCoordinateChangesImg(rc: TRect; minC, maxC: Vector2; preserveRatio: Boolean);
{Sets coefficients for the functions 'FlatImg3DToPanel' and its inverse}
var
  i,w, cut: Integer;
  edges: Vector2;
begin
  w:= rc.right - rc.left -1;
  imgCtr.x:= (rc.right + rc.left) div 2;
  imgCtr.y:= (rc.top + rc.bottom) div 2;
  minImg:= minC;
  maxImg:= maxC;
  if MapR2R2.modeStImage then begin
    stsImg.sc:= 0.45*(w+1)/stsImg.r;
    stsImg.ctr:= imgCtr;
  end
  else begin
    for i:= 1 to 2 do begin
      edges[i]:= maxImg[i]- minImg[i];
      img2panelSc[i]:=  w/edges[i];
    end;
    if preserveRatio then begin {the zoom factor must be equal in both directions.
        So, we take the smaller of the two and reduce conveniently the
      rectangle where the drawing is to be done}
      if img2panelSc[1] < img2panelSc[2] then begin
        cut:= round(0.5*(w - edges[2]*img2panelSc[1]));
        img2panelSc[2]:= img2panelSc[1];
        SetRect(imgEffRct, rc.left, rc.Top + cut, rc.right, rc.bottom - cut);
      end
      else begin
        img2panelSc[1]:= img2panelSc[2];
        cut:= round(0.5*(w - edges[1]*img2panelSc[2]));
        SetRect(imgEffRct, rc.left+ cut, rc.Top, rc.right - cut, rc.bottom);
      end;
    end
    else
      imgEffRct:= rc;
    for i:= 1 to 2 do
      panel2ImgSc[i]:= 1/img2panelSc[i];
  end;
end;

procedure UpdateMaps;
{make maps assignations according with the user settings}
begin
  case mapR2R2.mt of
    Real: begin
        theF:= theFReal;
        theFdiff1:= theFRealDiff1;
        theJac:= theFRealJac;
        theJacGrad:= theFRealJacGrad;
      end;
    Complex: begin
        theF:= theCF;
        theFdiff1:= theCFDiff1;
        theJac:= theCFJac;
        theJacGrad:= TheCFJacGrad;
      end;
    else
      ;
  end;
  if mapR2R2.modeStImage then begin
    ImgR3ToPanel:= ImgR3ToPanelSt;
    ImgPanelToR2:=  ImgPanelToR2St;
    DomToImgR3:= DomToImgR3St;
    ImgToImgPanel:= ImgToImgPanelSt;
  end
  else begin
    ImgR3ToPanel:= ImgR3ToPanelFlat;
    ImgPanelToR2:=  ImgPanelToR2Flat;
    DomToImgR3:= DomToImgR3Flat;
    ImgToImgPanel:= ImgToImgPanelFlat;
  end;
  if mapR2R2.modeStDomain then begin
    DomR3ToPanel:= DomR3ToPanelSt;
    DomPanelToR2:= DomPanelToR2St;
    DomToDomPanel:= DomToDomPanelSt;
  end
  else begin
    DomR3ToPanel:= DomR3ToPanelFlat;
    DomPanelToR2:= DomPanelToR2Flat;
    DomToDomPanel:= DomToDomPanelFlat;
  end;
end;


end.
