
{-----------------------------------------------------------------------------
 Unit Name: OCComplexFunct
 Project Name : MapR2toR2
 Author:    Angel Montesinos Amilibia
 July 1999-- December 2008
 Purpose: Implements class TOCComplexFunction that reads a formula in
    some complex variables, parses it and builds an evaluator.
-----------------------------------------------------------------------------}
{

SingR2, Singularities of smooth maps from R^2 to R^2, of from the complex plane
     to the compllex plane.
Copyright (C) 1999-2008 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain
}

{For the parsing, see detailed documentation in the unit
   "OCFunctDDDoubleSafeIntegrals.pas" of the program Superficies.

 For the compilation of the function to machime code, see below}

UNIT OCComplexFunct;


INTERFACE
  uses
    ConstTypesAndGlobals;

TYPE
{  c_number = RECORD
    re, im : double;
  END;}

  TOCComplexFunction = CLASS(TObject)
  PRIVATE
    rawText : AnsiString; {the text of the function}
    data : ARRAY OF Vector2; {stores the constants that appear in the formula}
    functionCode : AnsiString; {the function opCode}
    PROCEDURE LeaveRoomInFPUStack(neededPlaces : Integer);
    PROCEDURE PopFromAppStack;
    PROCEDURE CodifyFunction;
    FUNCTION FactorizeText(targetString : AnsiString;
        VAR mayEndInFunction : Boolean) : AnsiString;
    procedure PutCodeInExeBlock(cod: AnsiString; var exeBlock: PChar);
  PUBLIC
    usedVars: AnsiString; {really appearing in rawText}
    V : PROCEDURE(VAR r : Vector2; CONST X : ARRAY OF Vector2);
    VBlock: PChar;
    CONSTRUCTOR Create;
    DESTRUCTOR Destroy; OVERRIDE;
    procedure ParseNewText(t, vars : AnsiString);
  END;

var
  theComplexF: TOCComplexFunction;


IMPLEMENTATION
USES
  SysUtils, Math, Windows;

CONST
  pi : double = 3.1415926535897932384;
  halfPi : double = 1.5707963267948966192313;
  twoPi : double = 6.283185307179586476925;
  recipTwoPi : double = 0.1591549430918953357689;
  recipPi : double = 0.3183098861837906715378;
  kMaxNumOfNames = 47; {The number of strings of length less}
                             {than 8 that the parser recognizes as}
                             {function names. For instance, c and}
                             {conj are both recognized as meaning "Conjugate".}
  upCaseLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  variables = 'PQRSTUVWXYZ';
  digits = '0123456789';

  eMaxNum = 1.0E30;
  eMinNum = 1.0E-30;
  eMaxExp = 100;

  kSizDoub = SizeOf(double);

  {char constants}
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';
  kLastChar = #$FF;

  {Code constants}

  c_zero_comp = #$D9#$EE#$DF#$F1;
    {4 bytes. ST(7) must be empty. Compares 0 with x=ST(0). Then, JA jumps
     if 0>x; JNA jumps if 0 <= x; JB, if 0 < x; JNB, if 0 >= x; JNE if 0<>x}
  c_1_comp = #$D9#$E8#$DF#$F1;
    {4 bytes.  ST(7) must be empty. Compares 1 with x=ST(0). Then, JA jumps if
     1>x; JNA jumps if 1 <= x; JB, if 1 < x; JNB, if 1 >= x; JNE if 1<>x}
  c_mod_2pi = #$D9#$EB#$D8#$C0#$D9#$C9#$D9#$F8#$DD#$D9;
    {10 bytes. ST(cLastReg) must be empty.  changes st(0) mod 2pi  by st(0)}
  c_mult =  {x in st0), y in st(1), a in st(2), b in st(3)}
       #$D9#$C0  {fld st(0); x in st(0) and st(1), y in st(2), a in st(3), b in st(4)}
     + #$D8#$CB  {fmul st(0) st(3); ax in st(0), x in st(1), y in st(2), a in st(3), b in st(4)}
     + #$D9#$CB  {fxch st(3); a in st(0), x in st(1), y in st(2), ax in st(3), b in st(4)}
     + #$D8#$CA  {fmul st(0) st(2); ay in st(0), x in st(1), y in st(2), ax in st(3), b in st(4)}
     + #$D9#$CC  {fxch st(4); b in st(0), x in st(1), y in st(2), ax in st(3), ay in st(4)}
     + #$DC#$CA  {fmul st(2) st(0); b in st(0), x in st(1), by in st(2), ax in st(3), ay in st(4)}
     + #$DE#$C9  {fmulp  bx in st(0), by in st(1), ax in st(2), ay in st(3)}
     + #$DE#$C3  {faddp st(3) st(0); by in st(0), ax in st(1), bx + ay in st(2)}
     + #$DE#$E9; {fsubp; ax-by in st(0), bx + ay in st(1)}

  c_ln = { ln(st(0)) }
     #$D9#$E8#$D9#$C9#$D9#$F1#$D9#$EA#$DE#$F9;
       {10 bytes. Requires ST(0) > 0 and ST(7) empty}

  c_2toX =  {gives 2^st(0)}
       #$D9#$C0  {fld st(0); x in st(0) and st(1)}
     + #$D9#$FC  {frndint; int(x) in st(0), x in st(1)}
     + #$D9#$C9  {fxch; x in st(0=, int(x) in st(1)}
     + #$D8#$E1  {fsub st(0) st(1); x-int(x) in st(0), int(x) in st(1)}
     + #$D9#$F0  {f2xm1; 2^(frac(x))-1 in st(0), int(x) in st(1)}
     + #$D9#$E8  {fld1; 1 in st(0), 2^(frac(x))-1 in st(1), int(x) in st(2)}
     + #$DE#$C1  {faddp; 2^(frac(x)) in st(0), int(x) in st(1)}
     + #$D9#$C9  {fxch; int(x) in st(0), 2^(frac(x)) in st(1)}
     + #$D9#$E8  {fld1; 1 in st(0), int(x) in st(1), 2^(frac(x)) in st(2)}
     + #$D9#$FD  {fscale; 2^int(x) in st(0), int(x) in st(1), 2^(frac(x)) in st(2)}
     + #$DD#$D9  {fstp st(1); 2^int(x) in st(0), 2^(frac(x)) in st(1)}
     + #$DE#$C9; {fmulp; 2^x in st(0)}
    {24 bytes. Requires ST(6) and ST(7) empty}

  c_redAbsHalfPi = {if abs(st(0)) > pi/2, then
                        st(0):= pi-st(0)  and st(1):= -st(1)
                    16 bytes. Requires st(6) and st(7) empty   }
       #$D9#$C0  {fld st(0); a in st(0) and st(1), b in st(2)}
    +  #$D9#$E1  {fabs; abs(a) in st(0), a in st(1), b in st(2)}
    +  #$D8#$C0  {fadd st(0) st(0); 2*abs(a) in st(0), a in st(1), b in st(2)}
    +  #$D9#$EB  {fldpi; pi in st(0), 2*abs(a) in st(1), a in st(2), b in st(3)}
    +  #$D9#$C9  {fxch; 2*abs(a) in st(0), pi in st(1), a in st(2), b in st(3)}
    +  #$DF#$F1  {fcomip st(0) st(1); compares 2*abs(a) with pi and pop st(0); pi in st(0), a in st(1), b in st(2)}
    +  #$76#$18  {jbe; if 2*abs(a) <= pi, jump  $16=22 bytes }
    +  #$D9#$C9  {fxch; a in st(0), pi in st(1), b in st(2)}
    +  #$D9#$EE  {fldz; 0 in st(0), a in st(1), pi in st(2), b in st(3)}
    +  #$DF#$F1  {fcomip st(0) st(1); compares 0 with a and pop st(0); a in st(0), pi in st(1), b in st(2)}
    +  #$77#$04  {ja; if 0 > a, then jump $04 bytes}
    +  #$DE#$E9  {fsubp; pi-a in st(0), b in st(2)}
    +  #$EB#$04  {jump $04 bytes}
    +  #$DE#$C1  {faddp; pi+a in st(0), b in st(1)}
    +  #$D9#$E0  {fchs;  -pi-a in st(0), b in st(1)}
    +  #$D9#$C9  {fxch; b in st(0), pi-a or -pi-a in st(1)}
    +  #$D9#$E0  {fchs; -b in st(0), pi-a or -pi-a in st(1)}
    +  #$D9#$C9  {fxch; pi-a or -pi-a in st(0), -b in st(1)}
    +  #$EB#$02  {jump $02 bytes}
    +  #$DD#$D8;  {fstp st(0) st(0)}
    
  c_stackDisplacDouble = 8;  {Displacement in appl. stack for a double}
  c_stackDisplacDoubleHStr= #$08; {hexadecimal string of c_stackDisplacDouble}
  {c_stackDisplacExtended = 12;}
  {c_stackDisplacExtendedHStr= #$0C;} {hexadecimal string of c_stackDisplacExtended}

  c_LastReg = 7; {last register in the FPU}



TYPE

  pieceClass = (openParent, closeParent, operationSymb, {}
    negation, constantSymb, varSymb, functionSymb, badThing);

  WrittenFunction = RECORD
    theText : AnsiString;
    equivChar : char;
  END;



VAR
  uKnownFunc : ARRAY[1..kMaxNumOfNames] OF WrittenFunction;
  uFormula : AnsiString;
  uAttribution : ARRAY['#'..#$9F] OF pieceClass;
  uReduced : AnsiString;
  uSpecialValues : ARRAY[0..7] OF double;
  dataLast : byte;
  allowedVars : AnsiString; {variable names that may appear in 'rawText'.
                 must be a subset of [p,q,r,s,t,u,v,w,x,y,z]}
  argNum : Integer; {Length[allowedVars]}
  actorList : AnsiString; {the reverse Polish notation string that represents the function}
  lengthActorList : Integer;

  emptyRegisters, waitingInStack : Byte;

  {code vars}
  c_ln_S, c_2toX_S, c_sqrtZ, c_expZ, c_lnZ, c_sinZ, c_cosZ, c_tanZ, c_divZ,
    c_tanhZ, c_arcsinZ, c_arccosZ, c_arctanZ, c_atanhZ, c_invertZ :  AnsiString;



CONSTRUCTOR TOCComplexFunction.Create;
BEGIN
  INHERITED Create;
  rawText := '';
  functionCode := '';
  data := NIL;
END;

DESTRUCTOR TOCComplexFunction.Destroy;
BEGIN
  IF self <> NIL THEN
    BEGIN
      rawText := '';
      functionCode := '';
      data := NIL;
      if VBlock <> nil then begin
        VirtualFree(VBlock, 0, MEM_RELEASE);
        VBlock:= nil
      end;
    END;
  INHERITED Destroy;
END;



procedure UnifyChars;
VAR
  place, depth, anInteger, theLength : Integer;
  currentChar, precChar, lastAdded, toAdd : Char;
BEGIN {UnifyChars}
  theLength := Length(uFormula);
  uFormula := UpperCase(uFormula);
  uReduced := '';
  IF theLength = 0 THEN
    raise Exception.Create('Nothing written in formula');
  lastAdded := '(';
  currentChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  REPEAT
    precChar := currentChar;
    place := place + 1;
    currentChar := uFormula[place];
    CASE currentChar OF
      'A'..'Z' :
        BEGIN
          IF (lastAdded = ')') OR (pos(lastAdded, digits) <> 0) OR
            ((pos(lastAdded, upCaseLetters) <> 0) AND (precChar = ' ')) THEN
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        END;
      '0'..'9' :
        BEGIN
          IF (lastAdded = ')') OR (pos(lastAdded, upCaseLetters) <> 0) OR
            ((pos(lastAdded, digits) <> 0) AND (precChar = ' ')) THEN
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        END;
      '(', '[', '{' :
        BEGIN
          IF (lastAdded = ')') OR (pos(lastAdded, digits) <> 0) THEN
            uReduced := Concat(uReduced, '*');
          toAdd := '(';
          depth := depth + 1;
        END;
      '*' :
        IF (lastAdded = '*') THEN
          BEGIN
            anInteger := Length(uReduced);
            delete(uReduced, anInteger, 1);
            uReduced := concat(uReduced, '^');
            lastAdded := '^';
            toAdd := ' ';
          END
        ELSE
          toAdd := '*';
      ')', ']', '}' :
        BEGIN
          toAdd := ')';
          depth := depth - 1;
        END;
      ',', '.', '''' :
        toAdd := '.';
      '-' :
        IF (lastAdded = '(') THEN
          toAdd := '#'
        ELSE
          toAdd := '-';
      '/', '\' :
        toAdd := '/';
      ' ', '^' :
        toAdd := currentChar;
      '+' :
        IF (lastAdded = '(') THEN
          toAdd := ' '
        ELSE
          toAdd := '+';
      kReturn, kTab, kLineFeed :
        toAdd := ' ';
    ELSE
      raise Exception.Create('A bad character has been used');
    END; {of CASE}

    IF toAdd <> ' ' THEN
      BEGIN
        uReduced := concat(uReduced, toAdd);
        lastAdded := toAdd;
      END;
  UNTIL (place = theLength) OR (depth < 0);
  IF depth <> 0 THEN
    raise Exception.Create('Check please the parentheses');
  IF (place < theLength) THEN
    raise Exception.Create('Expression too long');
END; {of UnifyChars}

FUNCTION inSimpleNumber(targetStr : AnsiString;
  VAR theChar : char;
  VAR thePos : Integer) : Vector2;
  {targetString is any string. On start, it is assumed that theChar = }
  {targetStr[thePos] and that it is a digit.}
  {targetString is scanned until it ceases to give digits or dots. Only the following}
  {number formats are allowed: 1234,  1.234. Among formats not allowed are:}
  {.12, 1.2e-2, 12..34. If the number is not preceded by 'I*' nor followed}
  {by '*I' then the result is real; on the contrary, it is given as pure imaginary.}
  {On exiting, theChar is '?' if the number format was invalid;}
  {otherwise the function gives the number value, theChar is its last digit or 'I' and}
  {thePos is its position in targetStr.}
VAR
  valid, alreadyDot, foundOther, isImag : boolean;
  stringLength : Integer;
  antiguoSeparadorDecimal : char;
  numStr : AnsiString;
  temp : double;
BEGIN
  antiguoSeparadorDecimal := DecimalSeparator;
  DecimalSeparator := '.';
  stringLength := length(targetStr);
  alreadyDot := false;
  valid := true;
  numStr := theChar;
  foundOther := False;
  result[1] := 0;
  result[2] := 0;
  isImag := (thePos > 2) AND (targetStr[thePos - 1] = '*') AND (targetStr[thePos
    - 2] = 'I');
  WHILE (thePos < stringLength) AND valid AND (NOT foundOther) DO
    BEGIN
      inc(thePos);
      theChar := targetStr[thePos];
      CASE theChar OF
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' :
          numStr := numStr + theChar;
        '.' :
          IF alreadyDot THEN
            valid := False
          ELSE
            BEGIN
              alreadyDot := True;
              numStr := numStr + theChar;
            END;
      ELSE
        foundOther := true;
      END;
    END;
  IF NOT valid THEN
    theChar := '?'
  ELSE
    BEGIN
      IF isImag THEN
        result[2] := StrToFloat(numStr)
      ELSE
        result[1] := StrToFloat(numStr);
      IF foundOther THEN
        BEGIN
          IF (thePos + 1 <= stringLength) AND (targetStr[thePos] = '*') AND
            (targetStr[thePos + 1] = 'I') THEN
            BEGIN
              temp := result[2];
              result[2] := result[1];
              result[1] := -temp;
              thePos := thePos + 1;
              theChar := 'I';
            END
          ELSE
            BEGIN
              dec(thePos);
              theChar := targetStr[thePos];
            END;
        END;
    END;
  DecimalSeparator := AntiguoSeparadorDecimal;
END; {de inSimpleNumber}

FUNCTION inComplexNumber(targetStr : AnsiString;
  VAR theChar : char;
  VAR thePos : Integer) : Vector2;
  {targetString is any string, where capitalization does not matter. On start,
  {it is assumed that theChar = targetStr[thePos] and that it is a digit or 'I'.
  {It returns the value of the complex number that starts in thePos. Formats that
  are recognised are:  (4.34 + 3.21i), (4.34 + i 3.21), (4.34 i+ 3.21),  }
  {(i 4.34 + 3.21), i 4.34, 4.34 i,  2.34. On exit, theChar is the last}
  {character that forms the number and thePos, its position in targetStr.}
  {Note that in the case of a complex number with nonzero real and imaginary}
  {parts, the parentheses are considered as constitutive parts of the}
  {written number. If something goes wrong, theChar will be '?' on Exit}
VAR
  mayBeComplex, negative, keep : Boolean;
  stringLength, p: Integer;
  ch : char;
  secondVal : Vector2;
BEGIN
  result[1] := 0;
  result[2] := 0;
  secondVal:= result;
  stringLength := Length(targetStr);
  mayBeComplex := ((thePos > 1) AND (targetStr[thePos - 1] = '(')) OR ((thePos >
    3)
    AND (targetStr[thePos - 3] = '('));
  result := inSimpleNumber(targetStr, theChar, thePos);
  IF theChar = '?' THEN
    Exit;
  p := thePos;
  ch := theChar;
  keep := mayBeComplex AND (p + 2 < stringLength) AND (targetStr[p + 1] IN
    ['+', '-']);
  IF keep THEN
    BEGIN
      negative := (targetStr[p + 1] = '-');
      ch := targetStr[p + 2];
      IF (pos(ch, digits) <> 0) THEN
        BEGIN
          p := p + 2;
          secondVal := inSimpleNumber(targetStr, ch, p);
          keep := targetStr[p + 1] = ')';
        END
      ELSE IF (ch = 'I') AND (targetStr[p + 3] = '*') AND (p + 3 <
        stringLength) AND
        (pos(targetStr[p + 4], digits) <> 0) THEN
        BEGIN
          p := p + 4;
          ch := targetStr[p];
          secondVal := inSimpleNumber(targetStr, ch, p);
          keep := targetStr[p + 1] = ')';
        END
    END;
  IF keep THEN
    BEGIN
      IF negative THEN
        BEGIN
          secondVal[1] := -secondVal[1];
          secondVal[2] := -secondVal[2];
        END;
      theChar := ch;
      thePos := p;
      result[1] := result[1] + secondVal[1];
      result[2] := result[2] + secondVal[2];
    END;
END;

FUNCTION IdentifyFunction(theFunctionText : AnsiString) : char;
{If 'theFunctionText' is among the known functions, IdentifyFunction }
{returns the capital letter that corresponds to it. else, it    }
{returns '?'.                                                   }
VAR
  i : Integer;
  foundFunction : Boolean;
BEGIN
  i := 0;
  REPEAT
    i := i + 1;
    foundFunction := (theFunctionText = uKnownFunc[i].theText);
  UNTIL (i = kMaxNumOfNames) OR foundFunction;
  IF foundFunction THEN
    IdentifyFunction := uKnownFunc[i].equivChar
  ELSE
    IdentifyFunction := '?';
END;

FUNCTION TOCComplexFunction.FactorizeText (targetString : AnsiString; {}
  VAR mayEndInFunction : Boolean) : AnsiString;
VAR
  numberOfChars, hotSpot, count, position : Integer;
  currentChar : char;
  presumedText : AnsiString;
  resultingString : AnsiString;
BEGIN {FactorizeText}
  FactorizeText := '';
  resultingString := '';
  numberOfChars := Length(targetString);
  hotSpot := numberOfChars;
  IF mayEndInFunction THEN
    BEGIN
      count := numberOfChars;
      hotSpot := 1;
      REPEAT
        presumedText := Copy(targetString, hotSpot, count);
        currentChar := IdentifyFunction(presumedText);
        hotSpot := hotSpot + 1;
        count := count - 1;
      UNTIL (currentChar <> '?') OR (count = 0);
      IF (currentChar <> '?') THEN
        BEGIN
          hotSpot := hotSpot - 2;
          mayEndInFunction := TRUE;
          IF (hotSpot > 0) THEN
            resultingString := Concat('*', currentChar)
          ELSE
            BEGIN
              resultingString := currentChar;
              FactorizeText := resultingString;
              Exit;
            END;
        END
      ELSE
        BEGIN
          hotSpot := numberOfChars;
          mayEndInFunction := FALSE;
        END;
    END; {of IF mayEndInFunction ...}
  IF hotSpot > 0 THEN
    BEGIN
      currentChar := targetString[hotSpot];
      position := pos(currentChar, allowedVars);
      IF (position <> 0) THEN begin
        resultingString := Concat(kLastChar, chr(position), resultingString);
        if pos(currentChar, usedVars)=0 then
          usedVars:= usedVars + currentChar;
      end
      ELSE
        BEGIN {check if maybe it is a 'pi'}
          IF (currentChar = 'I') AND (hotSpot > 1) AND (targetString[hotSpot - 1]
            = 'P') THEN
            BEGIN
              hotSpot := hotSpot - 1;
              resultingString := 'p' + resultingString;
            END
          ELSE IF (currentChar = 'I') THEN
            resultingString := 'i' + resultingString
          ELSE {fatal error}
            Exit;
        END;
      hotSpot := hotSpot - 1;
    END;
  WHILE hotSpot > 0 DO
    BEGIN
      currentChar := targetString[hotSpot];
      position := pos(currentChar, allowedVars);
      IF (position <> 0) THEN begin
        resultingString := Concat(kLastChar, chr(position), '*',
          resultingString);
        if pos(currentChar, usedVars)=0 then
          usedVars:= usedVars + currentChar;
      end
      ELSE
        BEGIN {check if maybe it is a 'pi'}
          IF (currentChar = 'I') AND (hotSpot > 1) AND (targetString[hotSpot - 1]
            = 'P') THEN
            BEGIN
              hotSpot := hotSpot - 1;
              resultingString := 'p*' + resultingString;
            END
          ELSE IF (currentChar = 'I') THEN
            resultingString := 'i*' + resultingString
          ELSE {fatal error}
            Exit;
        END;
      hotSpot := hotSpot - 1;
    END; {of WHILE...}
  FactorizeText := resultingString;
END; {FactorizeText}


procedure ReduceSymbols(aF : TOCComplexFunction);
VAR
  numberOfChars, hotSpot, reducedLength : Integer;
  currentChar : char;
  letterStr : AnsiString;
  convertedStr, resultingString : AnsiString;
  numberBegins, inLetter, mayHaveFunction, foundEqualNumber : Boolean;
  numberValue : Vector2;
  aByte : Byte;
BEGIN {ReduceSymbols}
  resultingString := '';
  reducedLength := Length(uReduced);
  numberOfChars := reducedLength;
  hotSpot := 0;
  REPEAT
    hotSpot := hotSpot + 1;
    currentChar := uReduced[hotSpot];
    inLetter := (pos(currentChar, upCaseLetters) <> 0);
    numberBegins := (pos(currentChar, digits) <> 0) OR (currentChar = '.');
    IF (NOT inLetter) AND (NOT numberBegins) THEN
      ResultingString := Concat(resultingString, currentChar);
    IF numberBegins THEN
      BEGIN
        numberValue[1] := 0;
        numberValue[2] := 0;
        IF currentChar = '.' THEN
          currentChar := '?'
        ELSE
          numberValue := inComplexNumber(uReduced, currentChar, hotSpot);
        IF (currentChar <> '?') AND (currentChar <> '.') THEN
          BEGIN
            IF (numberValue[1] = 1) AND (numberValue[2] = 0) THEN
              resultingString := resultingString + '1'
            ELSE
              BEGIN
                foundEqualNumber := False;
                IF dataLast > 0 THEN
                  FOR aByte := 1 TO dataLast DO
                    IF (aF.data[aByte][1] = numberValue[1]) AND
                      (aF.data[aByte][2] = numberValue[2]) THEN
                      BEGIN
                        foundEqualNumber := True;
                        break;
                      END;
                IF foundEqualNumber THEN
                  resultingString := Concat(resultingString, kLastChar, chr(aByte
                    +
                    argNum))
                ELSE
                  BEGIN
                    inc(dataLast);
                    IF (dataLast + argNum = $FF - 1) THEN
                      raise Exception.Create('Too much constants in formula');
                    IF Length(aF.data) <= dataLast THEN
                      SetLength(aF.data, dataLast + 10);
                    aF.data[dataLast] := numberValue;
                    resultingString := Concat(resultingString, kLastChar,
                      chr(dataLast + argNum))
                  END;
              END
          END
        ELSE
          raise Exception.Create('A number was miswritten');
      END; {of numberBegins}
    IF inLetter THEN
      BEGIN
        letterStr := currentChar;
        WHILE (hotSpot < numberOfChars) AND inLetter DO
          BEGIN
            hotSpot := hotSpot + 1;
            currentChar := uReduced[hotSpot];
            inLetter := (pos(currentChar, upCaseLetters) <> 0);
            IF inLetter THEN
              letterStr := Concat(letterStr, currentChar);
          END;
        mayHaveFunction := (currentChar = '(');
        IF NOT inLetter THEN
          hotSpot := hotSpot - 1;
        convertedStr := aF.FactorizeText(letterStr, mayHaveFunction);
        IF convertedStr = '' THEN
          raise Exception.Create('Used a non-allowed variable or function');
        resultingString := Concat(resultingString, convertedStr);
        IF (currentChar = '(') AND (NOT mayHaveFunction) THEN
          ResultingString := Concat(resultingString, '*');
      END; {of inLetter}

  UNTIL (hotSpot >= numberOfChars);
  uReduced := resultingString;
END; {ReduceSymbols}

procedure CheckFunction(allowedIndepVars : AnsiString);
{Checks the sintaxis of the function}
VAR
  lastRead, reducedLength : integer;
  species, speciesAnt : pieceClass;
  charac : char;
  stopHere : boolean;

  FUNCTION hasError : boolean;
  VAR
    foundError, usual : boolean;
  BEGIN
    foundError := false;
    IF species = badThing THEN
      foundError := true
    ELSE
      BEGIN
        usual := (species = operationSymb) OR (species = closeParent);
        CASE speciesAnt OF
          openParent :
            foundError := usual;
          closeParent, constantSymb, varSymb :
            foundError := NOT usual;
          negation, operationSymb :
            foundError := usual OR (species = negation);
          functionSymb :
            foundError := (species <> openParent);
        END; {of case}
      END; {of else}
    hasError := foundError;
  END; {of hasError}

BEGIN {CheckFunction}
  lastRead := 0;
  reducedLength := Length(uReduced);
  speciesAnt := openParent;
  REPEAT
    lastRead := lastRead + 1;
    charac := uReduced[lastRead];
    IF charac = kLastChar THEN
      BEGIN
        inc(lastRead);
        species := constantSymb;
      END
    ELSE
      species := uAttribution[charac];
    stopHere := hasError;
    speciesAnt := species;
  UNTIL (lastRead = reducedLength) OR stopHere;
  IF (NOT stopHere) THEN
    BEGIN
      species := closeParent;
      stopHere := hasError;
    END;
  IF stopHere THEN
    raise Exception.Create('Syntax error in formula');
END; {CheckFunction}

FUNCTION ConvertPowerSymbol(x : AnsiString; n : Integer) : AnsiString;
VAR
  z : AnsiString;

  FUNCTION ConvertPowerSymbolProv(xx : AnsiString; nn : Integer) : AnsiString;
  VAR
    zz : AnsiString;
    m, np : Integer;
  BEGIN
    zz := xx;
    np := abs(nn);
    m := np DIV 2;
    IF np = 0 THEN
      Result := '1'
    ELSE IF nn = 1 THEN
      Result := xx
    ELSE IF m > 0 THEN
      BEGIN
        result := ConvertPowerSymbolProv(zz, m);
        result := 'B(' + result + ')';
        IF odd(np) THEN
          result := result + '*' + zz;
      END
    ELSE
      Result := zz;
  END;

BEGIN
  z := ConvertPowerSymbolProv(x, n);
  Result := '(' + z + ')';
  IF n < 0 THEN
    Result := '(' + '1/' + Result + ')';
END;

FUNCTION PieceEnd(CONST first : Integer; CONST theText, soughtChars :
  AnsiString; VAR symbol : char) : Integer;
{This function acts upon 'theText'. It is assumed that 'theText[first]'
is not in the set [+,-,*,/,^] (but possibly is '#'). Then, it scans forward
'theText' until in reaches one of the symbols in 'endChars' while being at
>=zero parentheses depth. Result is length[theText] if not found or the index
previous to that symbol.}
VAR
  i, len, depth : Integer;
  ch : char;
  found : Boolean;
BEGIN
  symbol := '?';
  len := Length(theText);
  i := first - 1;
  depth := 0;
  found := False;
  WHILE (i < len) AND NOT found DO
    BEGIN
      inc(i);
      ch := theText[i];
      IF (ch = '(') THEN
        inc(depth)
      ELSE IF (ch = ')') THEN
        dec(depth);
      IF ((pos(ch, soughtChars) <> 0) AND (depth = 0)) OR (depth < 0) THEN
        found := true;
    END;
  IF found THEN
    BEGIN
      PieceEnd := i - 1;
      symbol := ch;
    END
  ELSE
    PieceEnd := len;
END;

FUNCTION PieceBeginning(CONST last : Integer; soughtChars : AnsiString) :
  Integer;
{This function acts upon 'uReduced'. It is assumed that 'uReduced[last]'
is not in the set [+,-,*,/,^]. Then, it scans backwards
'uReduced' until in reaches one of the symbols in 'endChars' while being at
<=zero parentheses depth. Result is 1 (if not found) or the index following that symbol.}
VAR
  i, depth : Integer;
  ch : char;
  found : Boolean;
BEGIN
  i := last + 1;
  depth := 0;
  found := False;
  WHILE (i > 1) AND NOT found DO
    BEGIN
      dec(i);
      ch := uReduced[i];
      IF (ch = ')') THEN
        inc(depth)
      ELSE IF (ch = '(') THEN
        dec(depth);
      IF ((pos(ch, soughtChars) <> 0) AND (depth = 0)) OR (depth < 0) THEN
        found := true;
    END;
  IF found THEN
    PieceBeginning := i + 1
  ELSE
    PieceBeginning := 1;
END;

FUNCTION computeExponent(VAR n : Integer; aF : TOCComplexFunction; ini, fini :
  Integer) : Boolean;
VAR
  c1, c2 : char;
  negate : boolean;
  firstExp, lastExp, shift : Integer;
  exponent : double;
BEGIN
  firstExp := ini;
  lastExp := fini;
  c1 := uReduced[firstExp];
  c2 := uReduced[lastExp];
  IF (c1 = '(') THEN
    BEGIN
      result := (c2 = ')');
      IF result THEN
        BEGIN
          inc(firstExp);
          dec(lastExp);
          c1 := uReduced[firstExp];
          c2 := uReduced[lastExp];
        END;
    END
  ELSE
    result := (c1 = '1') OR (c1 = kLastChar);
  negate := (c1 = '#');
  IF negate THEN
    BEGIN
      inc(firstExp);
      c1 := uReduced[firstExp];
    END;

  IF (lastExp = firstExp) AND (c1 = '1') THEN
    BEGIN
      n := 1;
      Result := True;
    END
  ELSE IF ((lastExp - firstExp) = 1) AND (c1 = kLastChar) AND (ord(c2) > argNum)
    THEN
    BEGIN
      shift := ord(c2) - argNum;
      exponent := aF.data[shift][1];
      n := round(exponent);
      Result := (n = exponent) AND (aF.data[shift][2] = 0);
    END
  ELSE
    result := False;
  IF result AND negate THEN n := -n;
END;

FUNCTION convertPowers(aF : TOCComplexFunction) : Boolean;
VAR
  newString : AnsiString;
  reducedLength, firstBase, lastBase, firstExp, lastExp, i, n: Integer;
  c, c1 : char;
BEGIN
  reducedLength := Length(uReduced);
  result := true;
  IF reducedLength < 3 THEN
    Exit;
  i := 1;
  REPEAT
    inc(i);
    c := uReduced[i];
    Result := True;
    IF (c = '^') THEN
      BEGIN
        firstExp := i + 1;
        lastExp := PieceEnd(firstExp, uReduced, '+-*/^', c1);
        result := computeExponent(n, af, firstExp, lastExp);
        IF Result THEN
          BEGIN
            lastBase := i - 1;
            firstBase := PieceBeginning(lastBase, '+-*/^#');
            newString := Copy(uReduced, firstBase, lastBase - firstBase + 1);
            newString := ConvertPowerSymbol(newString, n);
            Insert(newString, uReduced, lastExp + 1);
            Delete(uReduced, firstBase, lastExp - firstBase + 1);
            reducedLength := Length(uReduced);
            i := firstBase + Length(newString) - 1;
          END
        ELSE
          Exit;
      END;
  UNTIL (i > reducedLength - 3);
END;



FUNCTION ExpressionString(theText : AnsiString) : AnsiString;
  FORWARD;

FUNCTION AtomString(theText : AnsiString) : AnsiString;
VAR
  base : AnsiString;
  len : Integer;
  ch : char;
BEGIN
  len := Length(theText);
  IF len = 0 THEN
    Result := ''
  ELSE
    BEGIN
      ch := theText[1];
      IF ch = '(' THEN
        BEGIN
          base := Copy(theText, 2, len - 2);
          result := ExpressionString(base);
        END
      ELSE
        BEGIN
          Assert(NOT (pos(ch, 'qrstuvwxyz') <> 0));
          CASE ch OF
            '1', 'i', 'p' :
              Result := ch;
            kLastChar :
              Result := theText;
            'A'..'Z' :
              BEGIN
                base := Copy(theText, 3, len - 3);
                result := ExpressionString(base) + ch;
              END;
          ELSE
            ;
          END;
        END;
    END;
END; {AtomString}

FUNCTION FactorString(theText : AnsiString) : AnsiString;
VAR
  neg : Boolean;
  redStr, anAtom : AnsiString;
  len, first, last : Integer;
  symbol, prevSymbol : char;
BEGIN
  IF (theText = '') THEN
    Result := ''
  ELSE
    BEGIN
      neg := (theText[1] = '#');
      IF neg THEN
        BEGIN
          len := Length(theText) - 1;
          redStr := Copy(theText, 2, len);
        END
      ELSE
        BEGIN
          len := Length(theText);
          redStr := theText;
        END;
      IF len = 0 THEN
        Result := ''
      ELSE
        BEGIN
          last := PieceEnd(1, redStr, '^', symbol);
          IF last = len THEN
            Result := AtomString(redStr)
          ELSE
            BEGIN
              anAtom := Copy(redStr, 1, last);
              prevSymbol := symbol;
              result := AtomString(anAtom);
              REPEAT
                first := last + 2;
                last := PieceEnd(first, redStr, '^', symbol);
                anAtom := Copy(redStr, first, last - first + 1);
                Result := AtomString(anAtom) + Result + prevSymbol;
                prevSymbol := symbol;
              UNTIL last = len;
            END;
          IF neg THEN Result := Result + '#';
        END;
    END;
END; {of FactorString}

FUNCTION TermString(theText : AnsiString) : AnsiString;
VAR
  aFactor : AnsiString;
  len, first, last : Integer;
  symbol, prevSymbol : char;
BEGIN
  len := Length(theText);
  IF len = 0 THEN
    Result := ''
  ELSE
    BEGIN
      last := PieceEnd(1, theText, '*/', symbol);
      IF last = len THEN
        Result := FactorString(theText)
      ELSE
        BEGIN
          aFactor := Copy(theText, 1, last);
          prevSymbol := symbol;
          result := FactorString(aFactor);
          REPEAT
            first := last + 2;
            last := PieceEnd(first, theText, '*/', symbol);
            aFactor := Copy(theText, first, last - first + 1);
            Result := Result + FactorString(aFactor) + prevSymbol;
            prevSymbol := symbol;
          UNTIL last = len;
        END;
    END;
END; {of TermString}


FUNCTION ExpressionString(theText : AnsiString) : AnsiString;
VAR
  aTerm : AnsiString;
  len, first, last : Integer;
  symbol, prevSymbol : char;
BEGIN
  len := Length(theText);
  IF len = 0 THEN
    Result := ''
  ELSE
    BEGIN
      last := PieceEnd(1, theText, '+-', symbol);
      IF last = len THEN
        Result := TermString(theText)
      ELSE
        BEGIN
          aTerm := Copy(theText, 1, last);
          prevSymbol := symbol;
          result := TermString(aTerm);
          REPEAT
            first := last + 2;
            last := PieceEnd(first, theText, '+-', symbol);
            aTerm := Copy(theText, first, last - first + 1);
            Result := Result + TermString(aTerm) + prevSymbol;
            prevSymbol := symbol;
          UNTIL last = len;
        END;
    END;
END; {of ExpressionString}

procedure TOCComplexFunction.ParseNewText(t, vars : AnsiString);
VAR
  i : Integer;
  ch : char;
BEGIN
  rawText := t;
  uFormula := t;
  usedVars:= '';
  IF vars = '' THEN
    BEGIN
      allowedVars := variables;
      argNum := Length(variables);
    END
  ELSE
    BEGIN
      allowedVars := vars;
      StrUpper(PChar(allowedVars));
      argNum := Length(vars);
      FOR i := 1 TO argNum DO
        BEGIN
          ch := allowedVars[i];
          IF NOT (pos(ch, variables) <> 0) THEN
            raise Exception.Create('Used a wrong variable');
        END;
    END;
  data := NIL;
  SetLength(data, 8);

  {data[0] is not used}
  data[0][1] := 0;
  data[0][2] := 0;

  dataLast := 0;
  UnifyChars;
  ReduceSymbols(self);
  CheckFunction(allowedVars);
  IF NOT ConvertPowers(self) THEN
    raise Exception.Create('Only integer exponents are allowed');
  actorList := expressionString(uReduced);
  lengthActorList := Length(actorList);
  CodifyFunction;
END;



function sn(n: Pointer): AnsiString;
{Converts a pointer to a string in the format that the CPU needs to
 read correctly a memory address}
var
  c: PChar;
  i: Integer;
begin
  c:= @n;
  Result := '';
  for i := 0 to SizeOf(Pointer) - 1 do
    Result := Result + c[i];
end;


PROCEDURE TOCComplexFunction.LeaveRoomInFPUStack(neededPlaces : Integer);
BEGIN
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  IF emptyRegisters = 0 THEN BEGIN
      functionCode := functionCode
       + #$D9#$F6 {fdecstp: ST(cLastReg] becomes ST(0)}
       + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
          +#$DD#$1C#$24;  {fstp [ESP]}
      inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then
    REPEAT
      functionCode := functionCode + #$D9 + chr($C0 + c_LastReg - emptyRegisters)
          {push last occupied ST to ST(0)}
      + #$83#$EC+ c_stackDisplacDoubleHStr
                                  {subtract c_stackDisplacDouble from ESP}
      +#$DD#$1C#$24;  {fstp [ESP]}
      functionCode := functionCode + #$DD + chr($C0 + c_LastReg - emptyRegisters);
         {last occupied place contains the former value: declare it empty}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
END;


PROCEDURE TOCComplexFunction.PopFromAppStack;
VAR
  m : byte;
BEGIN
  m := c_LastReg +1 - emptyRegisters;
  functionCode := functionCode + #$DD + chr($D8 + m) {copy ST(0) to ST(m), which
     is the first free register, and pop st(0)}

   + #$DD#$04#$24 {fld [ESP]}
   + #$83#$C4 + c_stackDisplacDoubleHStr {add c_stackDisplacDouble to ESP};
  {if using Extended type, use 'c_stackDisplacExtendedHStr' instead}

  functionCode := functionCode + #$D9 + chr($C8 + m); {exchange ST(m) with ST(0)}
  dec(emptyRegisters);
END;

procedure TOCComplexFunction.PutCodeInExeBlock(cod: AnsiString;
     var exeBlock: PChar);
var
  len: Integer;
  i: Integer;
  ba: ByteArray;
begin
  len:= Length(cod);
  SetLength(ba, len+2);
  for i:= 0 to len-1 do
    ba[i]:= ord(cod[i+1]);

  exeBlock:= VirtualAlloc(nil, len + SizeOf(exeBlock), MEM_COMMIT, PAGE_EXECUTE_READWRITE);
  for i:= 0 to len-1 do
    exeBlock[i]:= chr(ba[i]);
  ba:= nil;
end;

PROCEDURE TOCComplexFunction.CodifyFunction;
VAR
  i, j, n : Integer;
  ch : char;
  escape : Boolean;
BEGIN
  functionCode := '';
  emptyRegisters := 8;
  waitingInStack := 0;
  i := 0;
  WHILE i < lengthActorList DO
    BEGIN
      inc(i);
      ch := actorList[i];
      escape := (ch = kLastChar);
      IF escape OR (ch IN ['1', 'i', 'p']) THEN
        BEGIN
          IF escape THEN
            BEGIN
              inc(i);
              ch := actorList[i];
            END;
          LeaveRoomInFPUStack(2); {two registers will become occupied}
          IF escape THEN {ord(ch) denotes the order number  of
                          a constant or parameter in 'data'}
            BEGIN
              IF (ord(ch) > argNum) THEN
                BEGIN  {load constant or parameter from its position in 'data'}
                  functionCode := functionCode + #$DD#$05
                      + sn(@data[ord(ch) - argNum][2]);
                  functionCode := functionCode + #$DD#$05
                      + sn(@data[ord(ch) - argNum][1]);
                END
              ELSE
                BEGIN {load a variable from the address [EDX] displaced by a
                  quantity depending on the order number of the variable in
                  the list of arguments. Starting from [EDX], the compiler
                  furnishes the variable values on calling the evaluator}
                  n := ord(ch);
                  functionCode := functionCode + #$DD#$42 + chr(kSizDoub * (2 * n
                    - 1));
                  functionCode := functionCode + #$DD#$42 + chr(kSizDoub * (2 *
                    n - 2));
                END;
            END
          ELSE IF (ch = 'p') THEN  {load 'pi' as a real number. So, first load its
             imaginary part, that is zero, by the call #$D9#$EE, and then its real part
             by the call #$D9#$EB}
            functionCode := functionCode + #$D9#$EE#$D9#$EB

          ELSE IF (ch = '1') THEN {as before, load the constant 1}
            functionCode := functionCode + #$D9#$EE#$D9#$E8
          ELSE IF (ch = 'i') THEN {load 'i': that is load its imaginary part, which
              is 1, by means of #$D9#$E8 and then its real part, which is zero}
            functionCode := functionCode + #$D9#$E8#$D9#$EE;

          emptyRegisters := emptyRegisters - 2; {as a result of the preceding calls}
        END
      ELSE
        BEGIN
          IF (ch IN ['+', '-', '*', '/']) THEN
            BEGIN
              IF (emptyRegisters > 4) THEN
                FOR j := 5 TO emptyRegisters DO
                  PopFromAppStack;
              CASE ch OF
                '+' :
                  functionCode := functionCode {x in st0), y in st(1), a in st(2), b in st(3)}
                    + #$DE#$C2 {faddp st(2),st(0); y in st(0), x+a in st(1), b in st(2)}
                    + #$DE#$C2;{faddp st(2),st(0); x+a in st(0), y+b in st(1)}
                '-' :
                  functionCode := functionCode
                    + #$DE#$EA {fsubp st(2), st(0); y in st(0), a-x in st(1), b in st(2)}
                    + #$DE#$EA; {fsubp st(2), st(0); a-x in st(0), b-y in st(1)}
                '*' :
                  BEGIN
                    LeaveRoomInFPUStack(1);
                    functionCode := functionCode + c_mult;
                  END;
                '/' :
                  BEGIN
                    LeaveRoomInFPUStack(2);
                    functionCode := functionCode + c_divZ;
                  END;
              END;
              emptyRegisters := emptyRegisters + 2;
            END
          ELSE {opCode of a function}
            BEGIN
              IF (ch IN ['B', 'H']) THEN
                LeaveRoomInFPUStack(1)
              ELSE IF (ch IN ['C', 'I']) THEN
                LeaveRoomInFPUStack(2)
              ELSE IF (ch IN ['F', 'G', 'J', 'K']) THEN
                LeaveRoomInFPUStack(3)
              ELSE IF (ch IN ['L', 'M', 'N', 'O', 'P']) THEN
                LeaveRoomInFPUStack(4);
              CASE ch OF
                '#' :  {change sign}
                  functionCode := functionCode
                    + #$D9#$E0  {fchs; -a in st(0), b in st(1)}
                    + #$D9#$C9  {fxch; b sin st(0), -a in st(1)}
                    + #$D9#$E0  {fchs; -b in st(0), -a in st(1)}
                    + #$D9#$C9; {fxch; -a sin st(0), -b in st(1)}

                'A' :  {conjugate}
                  functionCode := functionCode
                    + #$D9#$C9  {fxch; b sin st(0), a in st(1)}
                    + #$D9#$E0  {fchs; -b in st(0), a in st(1)}
                    + #$D9#$C9; {fxch; a sin st(0), -b in st(1)}

                'B' :  {square}
                  functionCode := functionCode
                    + #$D9#$C0  {fld st(0); a in st(0) and st(1), b in st(2)}
                    + #$D8#$CA  {fmul st(0) st(1); a*b in st(0), a in st(1), b in st(2)}
                    + #$D8#$C0  {fadd st(0) st(0); 2*a*b in st(0), a in st(1), b in st(2)}
                    + #$D9#$CA  {fxch st(2); b in st(0), a in st(1), 2*a*b in st(2)}
                    + #$D8#$C8  {fmul st(0) st(0); b^2 in st(0), a in st(1), 2*a*b in st(2)}
                    + #$D9#$C9  {fxch; a in st(0), b^2 in st(1), 2*a*b in st(2)}
                    + #$D8#$C8  {fmul; st(0) st(0); a^2 in st(0), b^2 in st(1), 2*a*b in st(2)}
                    + #$DE#$E1; {fsubrp; a^2-b^2 in st(0), 2*a*b in st(1)}

                'C' : {square root}
                  functionCode := functionCode + c_sqrtZ;

                'D' : {module, that is mod(a+b*i) = sqrt(a^2+b^2)}
                  functionCode := functionCode
                    + #$D8#$C8  {fmul st(0) st(0); a^2 in st(0), b in st(1)}
                    + #$D9#$C9  {fxch; b in st(0), a^2 in st(1)}
                    + #$D8#$C8  {fmul st(0) st(0); b^2 in st(0), a^2 in st(1)}
                    + #$DE#$C1  {faddp; a^2+b^2 in st(0)}
                    + #$D9#$FA  {fsqrt; sqrt(a^2+b^2) in st(0)}
                    + #$D9#$EE  {fld0; 0 in st(0), sqrt(a^2+b^2) in st(1)}
                    + #$D9#$C9; {fxch; sqrt(a^2+b^2) in st(0), 0 in st(1)}

                'E' : {Argument, that is z = mod(z)*e^(i*arg(z))}
                  functionCode := functionCode
                    + #$D9#$F3  {fpatan; arctan(b/a) in st(0)}
                    + #$D9#$EE  {fld0; 0 in st(0), arctan(b/a) in st(1)}
                    + #$D9#$C9; {fxch; arctan(b/a) in st(1), o in st(1)}

                'F' : {sine}
                  functionCode := functionCode + c_sinZ;

                'G' : {cosine}
                  functionCode := functionCode + c_cosZ;

                'H' : {exponential}
                  functionCode := functionCode + c_expZ;

                'I' : {natural logarithm}
                  functionCode := functionCode + c_lnZ;

                'J' : { hyperbolic sine. sinh(z) = -i* sin(i*z) }
                  functionCode := functionCode
                    + #$D9#$C9  {fxch; b in st(0), a in st(1)}
                    + #$D9#$E0  {fchs; -b in st(0), a in st(1)}
                    + c_sinZ    {sin(i*z).re in st(0), sin(i*z).im in st(1)}
                    + #$D9#$E0  {fchs; -sin(i*z).re in st(0), sin(i*z).im in st(1)}
                    + #$D9#$C9; {fxch; sin(i*z).im in st(0), -sin(i*z).re in st(1)}

                'K' :  {hyperbolic cosine. cosh(z) = cos(i*z)}
                  functionCode := functionCode
                    + #$D9#$C9  {fxch; b in st(0), a in st(1)}
                    + #$D9#$E0  {fchs; -b in st(0), a in st(1)}
                    + c_cosZ;   {cos(i*z)}

                'L' :  {hyperbolic tangent}
                  functionCode := functionCode + c_tanhZ;

                'M' :  {tangent. tan(z) = -i*tanh(i*z)}
                  functionCode := functionCode
                    + c_tanZ;

                'N' :  {arcsin(z). arcsin(z)= -i*ln(i*z+ sqrt(1-z^2))}
                  functionCode := functionCode
                    + c_arcsinZ;

                'O' : {arccos(z), that is arccos(a+b*i) = -i*ln(z + sqrt(z^2-1))}
                  functionCode := functionCode
                    + c_arccosZ;
                    
                'P': {arctanh(z), that is arctan(h)= (1/2)*ln((1+z)/(1-z)) }
                  functionCode := functionCode
                    + c_atanhZ;

                'Q':   {argsh(z)= ln(z + sqrt(z^2+1)) = -i*arcsin(i*z)}
                  functionCode := functionCode
                    + #$D9#$C9  {fxch; b in st(0), a in st(1)}
                    + #$D9#$E0  {fchs; -b in st(0), a in st(1)}
                    + c_arcsinZ  { arcsin(i*z).re in st(0), arcsin(i*z).im in st(1)}
                    + #$D9#$E0  {fchs; -arcsin(i*z).re in st(0), arcsin(i*z).im in st(1)}
                    + #$D9#$C9;  {fxch;  arcsin(i*z).im in st(0), -arcsin(i*z).re in st(1)}

                'R':   {arccosh(z) =  i*arccos(z) }
                  functionCode := functionCode
                    + c_arccosZ { arccos(z).re in st(0), arccos(z).im in st(1)}
                    + #$D9#$C9  {fxch; arccos(z).im in st(0), arccos(z).re in st(1)}
                    + #$D9#$E0; {fchs; -arccos(z).im in st(0), arccos(z).re in st(1)}
                'S':  {arctan(z) = -i*atanh(i*z) }
                  functionCode := functionCode
                    + c_arctanZ;

                'T':  {cot(z) = 1/tan(z)}
                  functionCode := functionCode
                    + c_tanZ
                    + c_invertZ;

                'U':  {arccot(z)=  pi/2 - arctan(z)}
                  functionCode := functionCode
                    + c_arctanZ
                    + #$D9#$E0  {fchs; -atanZ.re in st(0), atanz.im in st(1)}
                    + #$D9#$C9  {fxch; atanZ.im in st(0), -atanz.re in st(1)}
                    + #$D9#$E0  {fchs; -atanZ.im in st(0), -atanz.re in st(1)}
                    + #$D9#$C9  {fxch; -atanZ.re in st(0), -atanz.im in st(1)}
                    + #$D9#$E8  {fld1; 1 in st(0), -atanZ.re in st(1), -atanz.im in st(2)}
                    + #$D9#$E0  {fchs; -1 in st(0), -atanZ.re in st(1), -atanz.im in st(2)}
                    + #$D9#$EB  {fldpi; pi in st(0), -1 in st(1), -atanZ.re in st(2), -atanz.im in st(3)}
                    + #$D9#$FD  {fscale; pi/2 in st(0), -1 in st(1), -atanZ.re in st(2), -atanz.im in st(3)}
                    + #$DC#$C2  {fadd st(2) st(0); pi/2 in st(0), -1 in st(1), cotZ.re in st(2), cotZ.im in st(3)}
                    + #$D9#$C9  {fxch; pi/2 in st(0), -1 in st(0), pi/2 in st(1), cotZ.re in st(2), cotZ.im in st(3)}
                    + #$DD#$D8 {pop; pi/2 in st(0), cotZ.re in st(1), cotZ.im in st(2)}
                    + #$DB#$F1  {fcomi st(0) st(1); pi/2 in st(0), cotZ.re in st(1), cotZ.im in st(2)}
                    + #$72#$0E  {jb; if pi/2 < cotZ.re, jumps $0E = 14 bytes}
                    + #$D9#$E0  {fchs; -pi/2 in st(0), cotZ.re in st(1), cotZ.im in st(2)}
                    + #$DB#$F1  {fcomi st(0) st(1); -pi/2 in st(0), cotZ.re in st(1), cotz.im in st(2)}
                    + #$76#$04  {jbe; if -pi/2 <= cotZ.re, jump $04 bytes}
                    + #$DC#$E9  {fsub st(0) st(1); -pi/2 in st(0), pi/2 + cotZ.re in st(1), cotz.im in st(2)}
                    + #$DC#$E9  {fsub st(0) st(1); -pi/2 in st(0), pi+cotZ.re in st(1), cotz.im in st(2)}
                    + #$DD#$D8  {pop; cotZ.re or pi+cotZ.re in st(0), cotZ.im in st(1) }
                    + #$EB#$04  {jump $04 bytes}
                    + #$DC#$E9  {fsub st(0) st(1); pi/2 in st(0), cotZ.re -pi/2 in st(1), cotZ.im in st(2)}
                    + #$DE#$E9;  {fsubp st(0) st(1); cotZ.re -pi in st(1), cotZ.im in st(2)}
              ELSE
                ;
              END;
            END;
        END;
    END; {del while}
  functionCode := functionCode + #$DD#$18#$DD#$58#$08#$C3;
  PutCodeInExeBlock(functionCode, VBlock);
  @V := @VBlock[0];
END;



PROCEDURE InitializeCodeVars;
BEGIN
  c_ln_S := {safe ln(st(0)) : for st(0) <=0, result is -eMaxExp}
       c_zero_comp
     + #$72#$0A  {jb; if 0 < x then jump 10 bytes}
     + #$DD#$05 + sn(@uSpecialValues[6]) {fld eMaxExp; eMaxExp in st(0), x in st(1)}
     + #$D9#$E0  {fchs; -eMaxExp in st(0), x in st(1)}
     + #$EB#$0A  {jump 10 bytes; so, this returns  -eMaxExp}
     + c_ln;
    {26 bytes. Requires ST(7) empty}

  c_2toX_S := {Safe 2^ST(0). If st(0) > eMaxExp, then result = eMaxNum;
               if st(0) < - eMaxExp, then  result = eMinNum}
       #$D9#$C0  {fld st(0); x in st(0) and st(1)}
     + #$D9#$E1  {fabs; abs(x) in st(0), s in st(1)}
     + #$DD#$05 + sn(@uSpecialValues[6]) {fld eMaxExp; eMaxExp in st(0), abs(x) in st(1), x in st(2)}
     + #$DE#$E9  {fsbup; abs(x)-eMaxExp in st(0), x in st(1)}
     + c_zero_comp
     + #$DD#$D8  {pop st(0); x in st(0)}
     + #$72#$1A  {jb; if 0 < abs(x)-eMaxExp, jump 26 bytes}
     + c_2toX    {2^x in st(0)}
     + #$EB#$16  {jump 22 bytes}
     + c_zero_comp {we know abs(x) >= eMaxExp; compares 0 with x}
     + #$DD#$D8  {all registers empty}
     + #$72#$08  {jb; if 0 < x then jump 8 bytes}
     + #$DD#$05 + sn(@uSpecialValues[4]) {eMinNum in st(0)}
     + #$EB#$06  {jump 6 bytes. Return eMinNum}
     + #$DD#$05 + sn(@uSpecialValues[2]); {fld eMaxNum; eMaxNum in st(0)}
     {72 bytes. Requires ST(6) and ST(7) empty.}

  c_sqrtZ:= #$D9#$C0 {fld st(0); a in st(0) and in st(1), b in st(2)}
     + #$D8#$C8  {fmul; a^2 in st(0), a in st(1), b in st(2)}
     + #$D9#$CA  {fxch st(2); b in st(0), a in st(1), a^2 in st(2)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), a in st(2), a^2 in st(3)}
     + #$D8#$C8  {fmul; b^2 in st(0), b in st(1), a in st(2), a^2 in st(3)}
     + #$DE#$C3  {faddp st(3) st(0); b in st(0), a in st(1), a^2+b^2 in st(2)}
     + #$D9#$CA  {fxch st(2); a^2+b^2 in st(0), a in st(1), b in st(2)}
     + #$D9#$FA  {fsqrt; sqrt(a^2+b^2) in st(0), a in st(1), b in st(2) }
     + #$D9#$C0  {fld st(0); sqrt(a^2+b^2) in st(0) and st(1), a in st(2), b in st(3)}
     + #$D8#$E2  {fsub st(0) st(2); sqrt(a^2+b^2)-a in st(0), sqrt(a^2+b^2) in st(1), a in st(2), b in st(3)}
     + #$D9#$CA  {fxch st(2); a in st(0), sqrt(a^2+b^2) in st(1), sqrt(a^2+b^2)-a in st(2), b in st(3)}
     + #$DE#$C1  {faddp; sqrt(a^2+b^2)+a in st(0), sqrt(a^2+b^2)-a in st(1), b in st(2)}
     + #$DD#$05 + sn(@uSpecialValues[7])  {fload 1/2;  1/2 in st(0), sqrt(a^2+b^2)+a in st(1), sqrt(a^2+b^2)-a in st(2), b in st(3)}
     + #$DC#$CA  {fmul st(2) st(0); 1/2 in st(0), sqrt(a^2+b^2)+a in st(1), (1/2)(sqrt(a^2+b^2)-a) in st(2), b in st(3)}
     + #$DE#$C9  {fmulp; (1/2)(sqrt(a^2+b^2)+a) in st(0), (1/2)(sqrt(a^2+b^2)-a) in st(1), b in st(2)}
     + #$D9#$FA  {fsqrt; sqrt((1/2)(sqrt(a^2+b^2)+a)) in st(0), (1/2)(sqrt(a^2+b^2)-a) in st(1), b in st(2)}
     + #$D9#$C9  {fxch; (1/2)(sqrt(a^2+b^2)-a) in st(0), sqrt((1/2)(sqrt(a^2+b^2)+a)) in st(1), b in st(2)}
     + #$D9#$FA  {fsqrt; sqrt((1/2)(sqrt(a^2+b^2)-a)) in st(0), sqrt((1/2)(sqrt(a^2+b^2)+a)) in st(1), b in st(2)}
     + #$D9#$CA  {fxch st(2); b in st(0), sqrt((1/2)(sqrt(a^2+b^2)-a)) in st(1), sqrt((1/2)(sqrt(a^2+b^2)+a)) in st(2)}
     + c_zero_comp {compares st(0) with 0}
     + #$DD#$D8  {fstp st(0); sqrt((1/2)(sqrt(a^2+b^2)-a)) in st(0), sqrt((1/2)(sqrt(a^2+b^2)+a)) in st(1)}
     + #$76#$02  {if b > 0 jump 2 bytes}
     + #$D9#$E0; {fchs; -sqrt((1/2)(sqrt(a^2+b^2)-a)) in st(0), sqrt((1/2)(sqrt(a^2+b^2)+a)) in st(1)}

  c_expZ :=      {    must give  exp(a+i*b)= exp(a)*(cos(b) + i sin(b))  }
       #$D9#$EA {fldl2e; log_2(e) in st(0), a in st(1), b in st(2)}
     + #$DE#$C9  {fmulp; a*log_2(e) in st(0), b in st(1)}
     + c_2ToX_S  {exp(a) in st(0), b in st(1)}
     + #$D9#$C9  {fxch; b in st(0), exp(a) in st(1)}
     + #$D9#$FB  {fsincos; cos(b) in st(0), sin(b) in st(1), exp(a) in st(2)}
     + #$D8#$CA  {fmul st(0) st(2); exp(a)*cos(b) in st(0), sin(b) in st(1), exp(a) in st(2)}
     + #$D9#$CA  {fxch st(2); exp(a) in st(0), sin(b) in st(1), exp(a)*cos(b) in st(2)}
     + #$DE#$C9  {fmulp; exp(a)*sin(b) in st(0), exp(a)*cos(b) in st(1)}
     + #$D9#$C9; {fxch; exp(a)*cos(b) in st(0), exp(a)*sin(b) in st(1)) }

  c_divZ :=  {must give  (a+i*b)/(x+i*y) =  (a*x +b*y + i*(-a*y+b*x))/(x^2+y^2)}
       #$D9#$C0  {fld st(0); x in st(0) and st(1), y in st(2), a in st(2), b in st(3)}
     + #$D8#$C8  {fmul; x^2 in st(0), x in st(1), y in st(2), a in st(2), b in st(3)}
     + #$D9#$CA  {fxch st(2); y in st(0), x in st(1), x^2 in st(2), a in st(2), b in st(3)}
     + #$D9#$E0  {fchs;  -y in st(0), x in st(1), x^2 in st(2), a in st(2), b in st(3)}
     + #$D9#$C0  {fld st(0);  -y in st(0) and st(1), x in st(2), x^2 in st(3), a in st(4), b in st(5)}
     + #$D8#$C8  {fmul;  y^2 in st(0), -y in st(1), x in st(2), x^2 in st(3), a in st(4), b in st(5)}
     + #$DE#$C3  {faddp st(3) st(0); -y in st(0), x in st(1), x^2+y^2 in st(2), a in st(3), b in st(4)}
     + #$D9#$CA  {fxch st(2); x^2+y^2 in st(0), x in st(1), -y in st(2), a in st(3), b in st(4)}
     + c_zero_comp
     + #$73#$18  {jae $18 bytes; jump $18 bytes if x^2+y^2 = 0}
     + #$DC#$FA  {fdiv st(2) st(0); x^2+y^2 in st(0), x in st(1), -y/(x^2+y^2) in st(2), a in st(3), b in st(4)}
     + #$DE#$F9  {fdivp; x/(x^2+y^2) in st(0), -y/(x^2+y^2) in st(1), a in st(2), b in st(3)}
     + c_mult    {occupies 18 bytes}{this gets the quotient we desired}
     + #$EB#$0E  {jump $0E bytes = 14 bytes, that is, terminates}
     + #$DD#$D8  {here we arrive from the jae $18 bytes = 24 bytes} {fstp st(0); x in st(0),
                               -y in st(1), a in st(2), b in st(3)}
     + #$DE#$D9  {fcompp; a in st(0), b in st(1)}
     + #$DD#$05 + sn(@uSpecialValues[2]) {fld m64 real; eMaxNum in st(0), a in st(1), b in st(2)}
     + #$DC#$CA  {fmul st(2) st(0); eMaxNum in st(0), a in st(1), b*eMaxNum in st(2)}
     + #$DE#$C9; {fmulp; a*eMaxNum in st(0), b*eMaxNum in st(1)}

  c_lnZ :=  {must give   ln(a+i*b) = (1/2)*ln(a^2+b^2) + i*Arg(a+i*b)}
       #$D9#$C0  {fld st(0); a in st(0) and st(1), b in st(1) }
     + #$D8#$C8  {fmul; a^2 in st(0), a in st(1), b in st(2)}
     + #$D9#$CA  {fxch  st(2); b in st(0), a in st(1), a^2 in st(2)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), a in st(2), a^2 in st(3)}
     + #$D8#$C8  {fmul; b^2 in st(0), b in st(1), a in st(2), a^2 in st(3)}
     + #$DE#$C3  {faddp st(3) st(0); b in st(0), a in st(1), a^2+b^2 in st(2)}
     + #$D9#$C9  {fxch; a in st(0), b in st(1), a^2+b^2 in st(2)}
     + #$D9#$F3  {fpatan; arctan(b/a) in st(0), a^2+b^2 in st(1)}
     + #$D9#$C9  {fxch; a^2+b^2 in st(0), arctan(b/a) in st(1)}
     +  c_ln_S   {ln(a^2+b^2) in st(0), arctan(b/a) in st(1)}
     + #$DD#$05 + sn(@uSpecialValues[7]) {load 1/2; 1/2 in st(0), ln(a^2+b^2) in st(1), arctan(b/a) in st(2)}
     + #$DE#$C9; {fmulp; (1/2)ln(a^2+b^2) in st(0), arctan(b/a) in st(1)}

  c_sinZ := {must give  sin(a+i*b)= sin(a) cosh(b)+i*cos(a)*sinh(b)}
       #$D9#$FB  {fsincos; cos(a) in st(0), sin(a) in st(1), b in st(2)}
     + #$D9#$CA  {fxch st(2); b in st(0), sin(a) in st(1), cos(a) in st(2)}
     + #$D9#$EA  {fldl2e; log_2(e) in st(0), b in st(1), sin(a) in st(2), cos(a) in st(3)}
     + #$DE#$C9  {fmulp; b*log_2(e) in st(0), sin(a) in st(1), cos(a) in st(2)}
     + c_2ToX_S  {exp(b) in st(0), sin(a) in st(1), cos(a) in st(2)}
     + #$D9#$E8  {fld1; 1 in st(0), exp(b) in st(1), sin(a) in st(2), cos(a) in st(3)}
     + #$D8#$F1  {fdiv st(0) st(1); exp(-b) in st(0), exp(b) in st(1), sin(a) in st(2), cos(a) in st(3)}
     + #$D9#$C0  {fld st(0); exp(-b) in st(0) and st(1), exp(b) in st(2), sin(a) in st(3), cos(a) in st(4)}
     + #$D8#$C2  {fadd st(0) st(2); exp(-b)+exp(b) in st(0), exp(-b) in st(1), exp(b) in st(2), sin(a) in st(3), cos(a) in st(4)}
     + #$DE#$CB  {fmulp st(3) st(0); exp(-b) in st(0), exp(b) in st(1), 2 cosh(b)*sin(a) in st(2),cos(a) in st(3)}
     + #$DE#$E9  {fsubp; exp(b)-exp(-b) in st(0), 2 cosh(b)*sin(a) in st(1), cos(a) in st(3)}
     + #$DE#$CA  {fmulp st(2) st(0); 2 sin(a)*cosh(b) in st(0),2 cos(a)*sinh(b) in st(1)}
     + #$DD#$05 + sn(@uSpecialValues[7]) {load 1/2; 1/2 in st(0), 2 sin(a)*cosh(b) in st(1),2 cos(a) sinh(b) in st(2)}
     + #$DC#$CA  {fmul st(2) st(0); 1/2 in st(0), 2 sin(a)*cosh(b) in st(1),cos(a)*sinh(b) in st(2)}
     + #$DE#$C9; {fmulp; sin(a)*cosh(b) in st(0),cos(a)*sinh(b) in st(1)}

  c_cosZ := {must give   cos(a+i*b)= cos(a) cosh(b)- i*sin(a) sinh(b)}
                               {the code is similar to that of c_sinZ}
       #$D9#$FB#$D9#$CA#$D9#$EA#$DE#$C9
     + c_2ToX_S + #$D9#$E8#$D8#$F1#$D9#$C0#$D8#$C2#$DE#$CC#$DE#$E1
     + #$DE#$C9#$DD#$05 + sn(@uSpecialValues[7])
     + #$DC#$CA#$DE#$C9#$D9#$C9;

  c_tanhZ := {  must give  (exp(2z)-1)/(exp(2z)+1)  }
       #$D8#$C0  {fadd st(0) st(0); 2*a in st(0), b in st(1)}
     + #$D9#$C9  {fxch; b in st(0), 2*a in st(1)}
     + #$D8#$C0  {fadd st(0) st(0); 2*b in st(0), 2*a in st(1)}
     + #$D9#$C9  {fxch; 2*a in st(0), 2*b in st(1)}
     + c_expZ   {exp(2z).re in st(0), exp(2z).im in st(1)}
     + #$D9#$C0  {fld st(0); exp(2z).re in st(0) and st(1), exp(2z).im in st(2)}
     + #$D9#$CA  {fxch st(2); exp(2z).im in st(0), exp(2z).re in st(1), exp(2z).re in st(2)}
     + #$D9#$C0  {fld st(0); exp(2z).im in st(0) and st(1), exp(2z).re in st(2), exp(2z).re in st(3)}
     + #$D9#$CB  {fxch st(3); exp(2z).re in st(0), exp(2z).im in st(1), exp(2z).re in st(2), exp(2z).im  in st(3)}
     + #$D9#$E8  {fld1; 1 in st(0), exp(2z).re in st(1), exp(2z).im in st(2), exp(2z).re in st(3), exp(2z).im  in st(4)}
     + #$DC#$C1  {fadd st(1) st(0); 1 in st(0), 1 + exp(2z).re in st(1), exp(2z).im in st(2), exp(2z).re in st(3), exp(2z).im  in st(4)}
     + #$DE#$EB  {fsubp st(3) st(0); 1 + exp(2z).re in st(0), exp(2z).im in st(1), exp(2z).re - 1 in st(2), exp(2z).im  in st(3)}
     + c_divZ;

  c_tanZ:=
       #$D9#$C9  {fxch; b in st(0), a in st(1)}
     + #$D9#$E0  {fchs; -b in st(0), a in st(1)}
     + c_tanhZ   {tanh(i*z).re in st(0), tanh(i*z).im in st(1)}
     + #$D9#$E0  {fchs; -tanh(i*z).re in st(0), tanh(i*z).im in st(1)}
     + #$D9#$C9; {fxch;  tanh(i*z).im in st(0), -tanh(i*z).re in st(1)}

   c_arcsinZ :=  {arcsin(z). arcsin(z)= -i*ln(i*z+ sqrt(1-z^2))}
       #$D9#$C0  {fld st(0); a in st(0) and st(1), b in st(2)}
     + #$D8#$C8  {fmul; a^2 in st(0), a in st(1), b in st(2)}
     + #$D9#$CA  {fxch st(2); b in st(0), a in st(1), a^2 in st(2)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), a in st(2), a^2 in st(3)}
     + #$D8#$C8  {fmul; b^2 in st(0), b in st(1), a in st(2), a^2 in st(3)}
     + #$DE#$E3  {fsubrp st(3) st(0); b in st(0), a in st(1), b^2-a^2 in st(2)}
     + #$D9#$E8  {fld1; 1 in st(0), b in st(1), a in st(2), b^2-a^2 in st(3)}
     + #$DE#$C3  {faddp st(3) st(0); b in st(0), a in st(1), 1+b^2-a^2 in st(2)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), a in st(2), 1+b^2-a^2 in st(3)}
     + #$D8#$CA  {fmul st(0) st(2); a*b in st(0), b in st(1), a in st(2), 1+b^2-a^2 in st(3)}
     + #$D9#$C0  {fld st(0); a*b in st(0) and st(1), b in st(2), a in st(3), 1+b^2-a^2 in st(4)}
     + #$DE#$C1  {faddp; 2*a*b in st(0), b in st(1), a in st(2), 1+b^2-a^2 in st(3)}
     + #$D9#$E0  {fchs; -2*a*b in st(0), b in st(1), a in st(2), 1+b^2-a^2 in st(3)}
     + #$D9#$C9  {fxch; b in st(0), -2*a*b in st(1), a in st(2), 1+b^2-a^2 in st(3)}
     + #$D9#$CB  {fxch st(3); 1+b^2-a^2 in st(0), -2*a*b in st(1), a in st(2), b in st(3)}
     + c_sqrtZ   {sqrt(1-z^2).re in st(0), sqrt(1-z^2).im in st(1), a in st(2), b in st(3)}
     + #$DE#$E3  {fsubrp st(3) st(0); sqrt(1-z^2).im in st(0), a in st(1), sqrt(1-z^2).re-b in st(2)}
     + #$DE#$C1  {faddp; a + sqrt(1-z^2).im in st(0), sqrt(1-z^2).re-b in st(1)}
     + #$D9#$C9  {fxch; sqrt(1-z^2).re-b in st(0), a + sqrt(1-z^2).im in st(1)}
     + c_lnZ     {ln(iz+sqrt(1-z^2)).re in st(0), ln(iz+sqrt(1-z^2)).im in st(1)}
     + #$D9#$E0  {fchs; -ln(iz+sqrt(1-z^2)).re in st(0), ln(iz+sqrt(1-z^2)).im in st(1)}
     + #$D9#$C9  {fxch; ln(iz+sqrt(1-z^2)).im in st(0),  -ln(iz+sqrt(1-z^2)).re in st(1)}
     + c_redAbsHalfPi;

   c_arccosZ:= {arccos(z). arccos(z)= i*ln(z+ sqrt(z^2-1))}
       #$D9#$C0  {fld st(0); a in st(0) and st(1), b in st(2)}
     + #$D8#$C8  {fmul; a^2 in st(0), a in st(1), b in st(2)}
     + #$D9#$CA  {fxch st(2); b in st(0), a in st(1), a^2 in st(2)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), a in st(2), a^2 in st(3)}
     + #$D8#$C8  {fmul; b^2 in st(0), b in st(1), a in st(2), a^2 in st(3)}
     + #$DE#$EB  {fsubp st(3) st(0); b in st(0), a in st(1), a^2-b^2 in st(3)}
     + #$D9#$E8  {fld1; 1 in st(0), b in st(1), a in st(2), a^2-b^2 in st(3)}
     + #$DE#$EB  {fsubp st(3) st(0); b in st(0), a in st(1), a^2-b^2-1 in st(3)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), a in st(2), a^2-b^2-1 in st(3)}
     + #$D8#$CA  {fmul st(0) st(2); a*b in st(0), b in st(1), a in st(2), a^2-b^2-1 in st(3)}
     + #$D9#$C0  {fld st(0); a*b in st(0) and st(1), b in st(2), a in st(3), a^2-b^2-1 in st(4)}
     + #$DE#$C1  {faddp; 2*a*b in st(0), b in st(1), a in st(2), a^2-b^2-1in st(3)}
     + #$D9#$C9  {fxch; b in st(0), 2*a*b in st(1), a in st(2), a^2-b^2-1 in st(3)}
     + #$D9#$CB  {fxch st(3); a^2-b^2-1 in st(0), 2*a*b in st(1), a in st(2), b in st(3)}
     + c_sqrtZ   {sqrt(z^2-1).re in st(0), sqrt(z^2-1).im in st(1), a in st(2), b in st(3)}
     + #$DE#$C2  {faddp st(2) st(0); sqrt(z^2-1).im in st(0), a +sqrt(z^2-1).re in st(1), b in st(2)}
     + #$DE#$C2  {faddp st(2) st(0); a +sqrt(z^2-1).re in st(0), sqrt(z^2-1).im in st(0)+ b in st(1)}
     + c_lnZ     {ln(z+sqrt(z^2-1)).re in st(0), ln(z+sqrt(z^2-1)).im in st(1)}
     + #$D9#$E0  {fchs; -ln(z+sqrt(z^2-1)).re in st(0), ln(z+sqrt(z^2-1)).im in st(1)}
     + #$D9#$C9  {fxch; ln(z+sqrt(z^2-1)).im in st(0), -ln(z+sqrt(z^2-1)).re in st(1)}
     + c_mod_2pi {of the real part of the result}
     + c_zero_comp
     + #$76#$08  {jbe; if st(0) >= 0 then jumps $04 bytes}
     + #$D9#$E0  {fchs; -result.re in st(0), result.im in st(1)}
     + #$D9#$C9  {fxch;  result.im in st(0), -result.re in st(1)}
     + #$D9#$E0  {fchs; -result.im in st(0), -result.re in st(1)}
     + #$D9#$C9; {fxch; -result.re in st(0), -result.im in st(1)}

   c_atanhZ:=
       #$D9#$C0  {fld st(0); a in st(0) and st(1), b in st(2)}
     + #$D9#$E8  {fld1; 1 in st(0), a in st(1) and st(2), b in st(3)}
     + #$DC#$C2  {fadd st(2) st(0); 1 in st(0), a in st(1), 1+a in st(2), b in st(3)}
     + #$DE#$E1  {fsubrp; 1-a in st(0), 1+a in st(1), b in st(2)}
     + #$D9#$CA  {fxch st(2); b in st(0), 1+a in st(1), 1-a in st(2)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), 1+a in st(2), 1-a in st(3)}
     + #$D9#$E0  {fchs; -b in st(0), b in st(1), 1+a in st(2), 1-a in st(3)}
     + #$D9#$C9  {fxch; b in st(0), -b in st(1), 1+a in st(2), 1-a in st(3)}
     + #$D9#$CB  {fxch st(3); 1-a in st(0), -b in st(1), 1+a in st(2), b in st(3)}
     + c_divZ    {(1+z)/(1-z).re in st(0), (1+z)/(1-z).im in st(1)}
     + c_lnZ     {ln((1+z)/(1-z)).re in st(0), ln((1+z)/(1-z)).im in st(1)}
     + #$DD#$05 + sn(@uSpecialValues[7])  {load 1/2}
     + #$DC#$CA  {fmul st(2) st(0); 1/2 in st(0), ln((1+z)/(1-z)).re in st(1),(1/2)*ln((1+z)/(1-z)).im in st(1) }
     + #$DE#$C9; {fmulp; (1/2)*ln((1+z)/(1-z)).re in st(0), (1/2)*ln((1+z)/(1-z)).im in st(1)}

   c_invertZ:= {1/z}
       #$D9#$C0  {fld st(0); a in st(0) and st(1), b in st(2)}
     + #$D8#$C8  {fmul; a^2 in st(0), a in st(1), b in st(2)}
     + #$D9#$CA  {fxch st(0) st(2); b in st(0), a in st(1), a^2 in st(2)}
     + #$D9#$C0  {fld st(0); b in st(0) and st(1), a in st(2), a^2 in st(3)}
     + #$D8#$C8  {fmul; b^2 in st(0), b in st(1), a in st(2), a^2 in st(3)}
     + #$DE#$C3  {faddp st(3) st(0); b in st(0), a in st(1), a^2+b^2 in st(2)}
     + #$D9#$E0  {fchs; -b in st(0), a in st(1), a^2+b^2 in st(2)}
     + #$D9#$CA  {fxch st(0) st(2), a^2+b^2 in st(0), a in st(1), -b in st(2)}
     + c_zero_comp
     + #$73#$06  {jae $06 bytes; jump $06 bytes if a^2+b^2 = 0}
     + #$DC#$F9  {fdiv st(1) st(0), a^2+b^2 in st(2), a/(a^2+b^2) in st(1), -b in st(2)}
     + #$DE#$FA  {fdivp st(2) st(0); a/(a^2+b^2) in st(0), -b/(a^2+b^2) in st(2) }
     + #$EB#$0C  {jump $0C = 12 bytes}
     + #$DD#$D8  {fstp st(0) st(0); a in st(0), -b in st(1)}
     + #$DD#$05 + sn(@uSpecialValues[2]) {fld m64 real; eMaxNum in st(0), a in st(1), -b in st(2)}
     + #$DC#$CA  {fmul st(2) st(0); eMaxNum in st(0), a in st(1), -b*eMaxNum in st(2)}
     + #$DE#$C9; {fmulp; a*eMaxNum in st(0), -b*eMaxNum in st(1)}

   c_arctanZ :=  {arctan(z) = -i*atanh(i*z) }
       #$D9#$C9  {fxch; b in st(0), a in st(1)}
     + #$D9#$E0  {fchs; -b in st(0), a in st(1)}
     + c_atanhZ  { arcsin(i*z).re in st(0), arcsin(i*z).im in st(1) }
     + #$D9#$E0  {fchs; -arcsin(i*z).re in st(0), arcsin(i*z).im in st(1)}
     + #$D9#$C9; {fxch;  arcsin(i*z).im in st(0), -arcsin(i*z).re in st(1)}

END;

PROCEDURE InitializeParser;
VAR
  character : char;
BEGIN

{strings that are recognized by the parser as functions and the char that identifies them}
  uKnownFunc[1].theText := 'C'; {Conjugation}
  uKnownFunc[1].equivChar := 'A';

  uKnownFunc[2].theText := 'CONJ'; {Conjugation}
  uKnownFunc[2].equivChar := 'A';

  uKnownFunc[3].theText := 'SQR'; {Square}
  uKnownFunc[3].equivChar := 'B';

  uKnownFunc[4].theText := 'SQRT'; {Square Root. But note that our implementation
  chooses always the value with non negative imaginary part, so that it complies with
  the mod*e^(i*arg) representation that would produce sqrt(mod)*e^(i*arg/2).}
  uKnownFunc[4].equivChar := 'C';

  uKnownFunc[5].theText := 'ABS'; {Module, that is abs(a+b*i) = sqrt(a^2+b^2)}
  uKnownFunc[5].equivChar := 'D';

  uKnownFunc[6].theText := 'MOD'; {Module, that is mod(a+b*i) = sqrt(a^2+b^2)}
  uKnownFunc[6].equivChar := 'D';

  uKnownFunc[7].theText := 'ARG'; {Argument, that is z = mod(z)*e^(i*arg(z))}
  uKnownFunc[7].equivChar := 'E';

  uKnownFunc[8].theText := 'SIN';
    {Sine, that is sin(a+b*i) = sin(a)*cosh(b)+ i* cos(a)*sinh(b)}
  uKnownFunc[8].equivChar := 'F';

  uKnownFunc[9].theText := 'SEN';
    {Sine, that is sin(a+b*i) = sin(a)*cosh(b)+ i* cos(a)*sinh(b)}
  uKnownFunc[9].equivChar := 'F';

  uKnownFunc[10].theText := 'COS';
    {Cosine, that is cos(a+b*i) = cos(a)*cosh(b)- i* sin(a)*sinh(b)}
  uKnownFunc[10].equivChar := 'G';

  uKnownFunc[11].theText := 'EXP';
  uKnownFunc[11].equivChar := 'H';

  uKnownFunc[12].theText := 'LN';
  uKnownFunc[12].equivChar := 'I';

  uKnownFunc[13].theText := 'LOG';
  uKnownFunc[13].equivChar := 'I';

  uKnownFunc[14].theText := 'SH';
  uKnownFunc[14].equivChar := 'J';

  uKnownFunc[15].theText := 'SINH';
  uKnownFunc[15].equivChar := 'J';

  uKnownFunc[16].theText := 'CH';
  uKnownFunc[16].equivChar := 'K';

  uKnownFunc[17].theText := 'COSH';
  uKnownFunc[17].equivChar := 'K';

  uKnownFunc[18].theText := 'TH';
  uKnownFunc[18].equivChar := 'L';

  uKnownFunc[19].theText := 'TANH';
  uKnownFunc[19].equivChar := 'L';

  uKnownFunc[20].theText := 'TAN';
  uKnownFunc[20].equivChar := 'M';

  uKnownFunc[21].theText := 'TG';
  uKnownFunc[21].equivChar := 'M';

  uKnownFunc[22].theText := 'ARCSEN';
  uKnownFunc[22].equivChar := 'N';

  uKnownFunc[23].theText := 'ARCSIN';
  uKnownFunc[23].equivChar := 'N';

  uKnownFunc[24].theText := 'ASEN';
  uKnownFunc[24].equivChar := 'N';

  uKnownFunc[25].theText := 'ASIN';
  uKnownFunc[25].equivChar := 'N';

  uKnownFunc[26].theText := 'ARCCOS';
  uKnownFunc[26].equivChar := 'O';

  uKnownFunc[27].theText := 'ARCOS';
  uKnownFunc[27].equivChar := 'O';

  uKnownFunc[28].theText := 'ACOS';
  uKnownFunc[28].equivChar := 'O';

  uKnownFunc[29].theText := 'ATANH';
  uKnownFunc[29].equivChar := 'P';

  uKnownFunc[30].theText := 'ARCTANH';
  uKnownFunc[30].equivChar := 'P';

  uKnownFunc[31].theText := 'ARGTH';
  uKnownFunc[31].equivChar := 'P';

  uKnownFunc[32].theText := 'ARGSH';
    {argsh(z)= ln(z + sqrt(z^2+1))}
  uKnownFunc[32].equivChar := 'Q';

  uKnownFunc[33].theText := 'ARGSINH';
  uKnownFunc[33].equivChar := 'Q';

  uKnownFunc[34].theText := 'ARCSINH';
  uKnownFunc[34].equivChar := 'Q';

  uKnownFunc[35].theText := 'ASINH';
  uKnownFunc[35].equivChar := 'Q';

  uKnownFunc[36].theText := 'ARGCH';
    {argch(z)= ln(z + sqrt(z^2-1))}
  uKnownFunc[36].equivChar := 'R';

  uKnownFunc[37].theText := 'ARGCOSH';
  uKnownFunc[37].equivChar := 'R';

  uKnownFunc[38].theText := 'ARCCOSH';
  uKnownFunc[38].equivChar := 'R';

  uKnownFunc[39].theText := 'ACOSH';
  uKnownFunc[39].equivChar := 'R';

  uKnownFunc[40].theText := 'ARCTAN';
  uKnownFunc[40].equivChar := 'S';

  uKnownFunc[41].theText := 'ARCTG';
  uKnownFunc[41].equivChar := 'S';

  uKnownFunc[42].theText := 'ATAN';
  uKnownFunc[42].equivChar := 'S';

  uKnownFunc[43].theText := 'COTG';
  uKnownFunc[43].equivChar := 'T';

  uKnownFunc[44].theText := 'COTAN';
  uKnownFunc[44].equivChar := 'T';

  uKnownFunc[45].theText := 'COT';
  uKnownFunc[45].equivChar := 'T';

  uKnownFunc[46].theText := 'ARCCOTG';
  uKnownFunc[46].equivChar := 'U';

  uKnownFunc[47].theText := 'ARCCOT';
  uKnownFunc[47].equivChar := 'U';



  {classification of symbols by their associated syntax}
  FOR character := '#' TO 'z' DO
    CASE character OF
      '(' :
        uAttribution[character] := openParent;
      '#' :
        uAttribution[character] := negation;
      ')' :
        uAttribution[character] := closeParent;
      '+', '-', '*', '/', '^' :
        uAttribution[character] := operationSymb;
      '1', 'p', 'i' :
        uAttribution[character] := constantSymb;
      'q'..'z' :
        uAttribution[character] := varSymb;
      'A'..'Z' :
        uAttribution[character] := functionSymb;
    ELSE
      uAttribution[character] := badThing;
    END; {of case}

  uSpecialValues[0] := halfPi;
  uSpecialValues[1] := twoPi;
  uSpecialValues[2] := eMaxNum;
  uSpecialValues[3] := -eMaxNum;
  uSpecialValues[4] := eMinNum;
  uSpecialValues[5] := -eMinNum;
  uSpecialValues[6] := eMaxExp;
  uSpecialValues[7] := 1/2;

  InitializeCodeVars;
  theComplexF:= TOCComplexFunction.Create;
END; {InitializeParser}

procedure FreeFunctions;
begin
  theComplexF.Free;
end;



INITIALIZATION

  InitializeParser;

finalization
  FreeFunctions;

END.

